"""Super droplet method for coagulation dynamics.

This module implements a Super Droplet Method for coagulation
dynamics, used to simulate how particles grow through collisions.
"""

from itertools import combinations_with_replacement
from typing import Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import RectBivariateSpline  # type: ignore


def _super_droplet_update_step(
    particle_radius: NDArray[np.float64],
    concentration: NDArray[np.float64],
    single_event_counter: NDArray[np.int64],
    small_index: NDArray[np.int64],
    large_index: NDArray[np.int64],
) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.int64]]:
    """Update particle radii and concentrations when two particles coagulate.

    This function merges smaller and larger particles by combining their
    volumes and redistributing particle concentrations. The resulting
    radii are computed via volume conservation, and an event counter
    tracks how many coagulation events each particle has undergone.

    Arguments:
        - particle_radius : Array of particle radii (m).
        - concentration : Array representing the concentration of each
          particle (number or mass, depending on usage).
        - single_event_counter : Tracks the number of coagulation events
          each particle has experienced in the current iteration.
        - small_index : Indices for smaller particles in a coagulation event.
        - large_index : Indices for larger particles in a coagulation event.

    Returns:
        - An updated array of particle radii (m) following coagulation.
        - An updated array representing the concentration of particles.
        - An updated array tracking the index-wise number of events.

    Examples:
        ```py
        import numpy as np
        r = np.array([1e-9, 2e-9, 3e-9])
        conc = np.array([100., 50., 75.])
        events = np.zeros_like(r, dtype=int)
        s_idx = np.array([0])
        l_idx = np.array([2])
        out_r, out_c, out_ev = _super_droplet_update_step(
            r, conc, events, s_idx, l_idx)
        # out_r[0] is updated via volume combination with out_r[2].
        ```
    """
    # Step 1: Calculate the summed volumes of the smaller and larger particles
    # The volumes are obtained by cubing the radii of the particles.
    sum_radii_cubed = np.power(
        particle_radius[small_index], 3, dtype=np.float64
    ) + np.power(particle_radius[large_index], 3, dtype=np.float64)

    # Step 2: Calculate the new radii formed by the coagulation events
    # The new radius is the cube root of the summed volumes.
    new_radii = np.cbrt(sum_radii_cubed)

    # Step 3: Determine the concentration differences between small and
    # large particles
    concentration_delta = (
        concentration[small_index] - concentration[large_index]
    )
    more_small = concentration_delta > 0
    more_large = concentration_delta < 0
    equal_concentration = concentration_delta == 0

    # Step 4: Handle cases where small and large particle concentrations are
    # equal. In these cases, split the concentrations equally and update
    # both small and large particle radii.
    if np.any(equal_concentration):
        # print("Handling equal concentration case (split)")
        concentration[small_index[equal_concentration]] /= 2
        concentration[large_index[equal_concentration]] /= 2

        particle_radius[small_index[equal_concentration]] = new_radii[
            equal_concentration
        ]
        particle_radius[large_index[equal_concentration]] = new_radii[
            equal_concentration
        ]

    # Step 5: Handle cases where there are more large particles than small ones
    # Update the concentration of large particles and adjust the radii of
    # small particles.
    if np.any(more_large):
        # print("Handling more large particles case")
        concentration[large_index[more_large]] = np.abs(
            concentration_delta[more_large]
        )
        particle_radius[small_index[more_large]] = new_radii[more_large]

    # Step 6: Handle cases where there are more small particles than large ones
    # Update the concentration of small particles and adjust the radii of
    # large particles.
    if np.any(more_small):
        # print("Handling more small particles case")
        concentration[small_index[more_small]] = np.abs(
            concentration_delta[more_small]
        )
        particle_radius[large_index[more_small]] = new_radii[more_small]

    # Increment event counters for both small and large particles
    single_event_counter[small_index] += 1
    single_event_counter[large_index] += 1

    return particle_radius, concentration, single_event_counter


def _event_pairs(
    lower_bin: int,
    upper_bin: int,
    kernel_max: Union[float, NDArray[np.float64]],
    number_in_bins: Union[NDArray[np.float64], NDArray[np.int64]],
) -> float:
    """Calculate an approximate count of particle-pair interactions.

    This function estimates the number of collisions or interactions
    that might occur between two bins of particles, given a maximum
    kernel value and the current population of each bin. When the bins
    are the same, a correction factor is applied to avoid double-counting
    pairs.

    Arguments:
        - lower_bin : Index of the lower bin in the distribution.
        - upper_bin : Index of the upper bin in the distribution.
        - kernel_max : Maximum kernel value used to weight collisions.
        - number_in_bins : The population of particles per bin.

    Returns:
        - A float representing the expected number of particle-pair
          collision events.

    Examples:
        ```py
        max_kernel = 1.0e-9
        n_bins = np.array([100, 150, 200])
        # lower_bin=0, upper_bin=1 => collisions between bin 0 and bin 1
        events_est = _event_pairs(0, 1, max_kernel, n_bins)
        # events_est is ~ 1.0e-9 * 100 * 150
        ```
    """
    # Calculate the number of particle pairs based on the kernel value
    if lower_bin != upper_bin:
        return (
            kernel_max * number_in_bins[lower_bin] * number_in_bins[upper_bin]
        )
    return (
        kernel_max
        # * 0.5
        * number_in_bins[lower_bin]
        * (number_in_bins[upper_bin] - 1)
    )


def _sample_events(
    events: float,
    volume: float,
    time_step: float,
    generator: np.random.Generator,
) -> int:
    """Determine how many collisions actually occur using a Poisson draw.

    This function uses the expected collision count (`events`) and normalizes
    by system `volume` to compute an effective collision rate. It then
    samples from a Poisson distribution to obtain the actual number of
    collisions happening within the current `time_step`.

    Arguments:
        - events : The calculated number of particle pairs that could
          interact.
        - volume : The volume of the simulation space (m³).
        - time_step : The time span (seconds) over which collisions are
          considered.
        - generator : A NumPy random Generator to sample the Poisson
          distribution.

    Returns:
        - The sampled number of coagulation events as an integer.

    Examples:
        ```py
        from numpy.random import default_rng
        rng = default_rng(42)
        collisions = _sample_events(events=5e3, volume=0.1, time_step=0.01,
            generator=rng)
        # collisions might be ~ Poisson( 5e3 / 0.1 * 0.01 ) => Poisson(5)
        ```
    """
    # Calculate the expected number of events
    events_exact = events / volume

    # Sample the actual number of events from a Poisson distribution
    return generator.poisson(events_exact * time_step)


# pylint: disable=too-many-positional-arguments, too-many-arguments
def random_choice_indices(
    lower_bin: int,
    upper_bin: int,
    events: int,
    particle_radius: NDArray[np.float64],
    bin_indices: NDArray[np.int64],
    generator: np.random.Generator,
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Select valid particle indices in two bins for coagulation events.

    This function tries to choose `events` valid indices from
    `lower_bin` and `upper_bin`, discarding any particles with radius ≤ 0.
    It uses the provided random generator to perform the sampling
    with replacement if needed.

    Arguments:
        - lower_bin : Index of the lower bin to filter particles from.
        - upper_bin : Index of the upper bin to filter particles from.
        - events : Number of events (indices) to sample for each bin.
        - particle_radius : Array of particle radii; only those > 0
          are considered valid.
        - bin_indices : Array of bin labels corresponding to each particle.
        - generator : Random number generator used for index selection.

    Returns:
        - Indices of particles from the lower bin.
        - Indices of particles from the upper bin.

    Examples:
        ```py
        import numpy as np
        rng = np.random.default_rng(123)
        radius = np.array([0.3, 0.1, 0.0, 0.5])
        bins = np.array([0, 0, 1, 1])
        lw_bin, up_bin = random_choice_indices(0, 1, 2, radius, bins, rng)
        # lw_bin -> array of valid picks from bin 0
        # up_bin -> array of valid picks from bin 1
        ```
    """
    try:
        # Directly find the indices where the condition is True
        lower_indices = generator.choice(
            np.flatnonzero((bin_indices == lower_bin) & (particle_radius > 0)),
            events,
            replace=True,
        )
        upper_indices = generator.choice(
            np.flatnonzero((bin_indices == upper_bin) & (particle_radius > 0)),
            events,
            replace=True,
        )
    except ValueError:
        # If no valid indices are found, return empty arrays
        return np.array([], dtype=np.int64), np.array([], dtype=np.int64)
    return lower_indices, upper_indices


def _select_random_indices(
    lower_bin: int,
    upper_bin: int,
    events: int,
    number_in_bins: NDArray[np.int64],
    generator: np.random.Generator,
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Randomly choose indices within each bin to represent collision partners.

    This function picks `events` indices from the population of the
    `lower_bin` and `upper_bin`, ignoring any radius or event-limit checks
    (those may happen later). The result is two arrays of equal size,
    each containing random picks within the respective bins.

    Arguments:
        - lower_bin : Index for the "smaller" bin.
        - upper_bin : Index for the "larger" bin.
        - events : How many pairs to select.
        - number_in_bins : Array with the count of particles in each bin.
        - generator : Random number generator to draw the indices.

    Returns:
        - An array of size `events` with random picks from `lower_bin`.
        - An array of size `events` with random picks from `upper_bin`.

    Examples:
        ```py
        import numpy as np
        rng = np.random.default_rng(42)
        n_in_bins = np.array([5, 10, 7])
        i_lw, i_up = _select_random_indices(
            lower_bin=0,
            upper_bin=2,
            events=3,
            number_in_bins=n_in_bins,
            generator=rng
        )
        # i_lw -> random indices in [0..4]
        # i_up -> random indices in [0..6]
    """
    # Select random indices for particles in the lower_bin
    lower_indices = generator.integers(  # type: ignore
        0,
        number_in_bins[lower_bin],
        size=events,
        endpoint=False,
        dtype=np.int64,
    )
    # Select random indices for particles in the upper_bin
    upper_indices = generator.integers(  # type: ignore
        0,
        number_in_bins[upper_bin],
        size=events,
        endpoint=False,
        dtype=np.int64,
    )
    return lower_indices, upper_indices


def _bin_to_particle_indices(
    lower_indices: NDArray[np.int64],
    upper_indices: NDArray[np.int64],
    lower_bin: int,
    upper_bin: int,
    bin_indices: NDArray[np.int64],
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Map bin-relative indices back to absolute positions in the particle
    array.

    This function adjusts the offsets for each bin so that the pairwise
    indices used for collision are mapped onto the actual sorted particle
    array. For instance, if `lower_indices` are all within bin 0, and bin 0
    particles occupy positions [0..9], this method adds that offset to
    each index in `lower_indices`.

    Arguments:
        - lower_indices : Relative indices (local to the bin) of smaller
            particles.
        - upper_indices : Relative indices (local to the bin) of larger
            particles.
        - lower_bin : The bin representing the smaller particles.
        - upper_bin : The bin representing the larger particles.
        - bin_indices : Cumulative offsets to determine where each bin begins.

    Returns:
        - `small_index` : Absolute positions of smaller particles in the
          sorted particle array.
        - `large_index` : Absolute positions of the larger particles in
          the sorted array.

    Examples:
        ```py
        bins = np.array([0, 10, 20])
        lw_rel = np.array([0, 1])
        up_rel = np.array([2, 3])
        # Convert these local indices for bin 1 (start=10) and bin 2 (start=20)
        s_idx, l_idx = _bin_to_particle_indices(lw_rel, up_rel, 1, 2, bins)
        # s_idx -> [10, 11]
        # l_idx -> [22, 23]
    """
    # Get the start index in the particle array for the lower_bin
    start_index_lower_bin = np.searchsorted(bin_indices, lower_bin)
    # Get the start index in the particle array for the upper_bin
    start_index_upper_bin = np.searchsorted(bin_indices, upper_bin)

    # Calculate the actual particle indices for the lower_bin and upper_bin
    small_index = start_index_lower_bin + lower_indices
    large_index = start_index_upper_bin + upper_indices

    return small_index, large_index


def _filter_valid_indices(
    small_index: NDArray[np.int64],
    large_index: NDArray[np.int64],
    particle_radius: NDArray[np.float64],
    single_event_counter: Optional[NDArray[np.int64]] = None,
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Remove invalid pairs of particles based on radius and optional event
    limit.

    This function checks each pair of `(small_index, large_index)` to ensure
    both have radius > 0. If `single_event_counter` is provided, it further
    enforces that each particle has had < 1 event so far (or you can
    define your own threshold). The pairs failing these checks are removed.

    Arguments:
        - small_index : Indices for the smaller particles in each pair.
        - large_index : Indices for the larger particles in each pair.
        - particle_radius : Array of radii for each particle.
        - single_event_counter : Optional array counting how many events
          each particle has undergone. If provided, only particles with
          counter < 1 pass the filter.

    Returns:
        - Filtered `small_index` with only valid pairs.
        - Filtered `large_index` that corresponds to valid pairs.

    Examples:
        ```py
        r = np.array([0.1, 0.0, 0.08, 0.02])
        c = np.array([0, 0, 0, 0])
        small_i = np.array([0, 1, 2])
        large_i = np.array([3, 0, 1])
        # Filter out pairs with radius <= 0 or event_counter >= 1
        s_valid, l_valid = _filter_valid_indices(
            small_i, large_i, r, single_event_counter=c
        )
        # Indices with r>0 remain in s_valid, l_valid
        ```
    """
    if single_event_counter is not None:
        # Both particle radius and event counter are used to determine
        # valid indices
        valid_indices = (
            (particle_radius[small_index] > 0)
            & (particle_radius[large_index] > 0)
            & (single_event_counter[small_index] < 1)
            & (single_event_counter[large_index] < 1)
        )
    else:
        # Only particle radius is used to determine valid indices
        valid_indices = (particle_radius[small_index] > 0) & (
            particle_radius[large_index] > 0
        )

    # Return the filtered indices
    return small_index[valid_indices], large_index[valid_indices]


def _coagulation_events(
    small_index: NDArray[np.int64],
    large_index: NDArray[np.int64],
    kernel_values: NDArray[np.float64],
    kernel_max: float,
    generator: np.random.Generator,
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Stochastically pick which collisions (among possible pairs) actually
    happen.

    This function computes a collision probability for each `(small_index,
    large_index)` pair by taking the ratio of `kernel_values / kernel_max`.
    Next, a random uniform draw decides if each collision occurs.

    Arguments:
        - small_index : Array of indices representing smaller particles.
        - large_index : Array of indices representing larger particles.
        - kernel_values : Collision kernel values for each pair.
        - kernel_max : A maximum kernel value used for normalization.
        - generator : Random generator to compare probabilities vs.
          uniform draws.

    Returns:
        - Filtered `small_index` containing only those that coagulated.
        - Filtered `large_index` containing only those that coagulated.

    Examples:
        ```py
        rng = np.random.default_rng(999)
        s_idx = np.array([0, 1, 2])
        l_idx = np.array([3, 4, 5])
        kv = np.array([0.5, 1.0, 0.1])
        kmax = 1.0
        s_new, l_new = _coagulation_events(s_idx, l_idx, kv, kmax, rng)
        # Each pair has probability kv/kmax => [0.5, 1.0, 0.1]
        # The final s_new, l_new depends on random draws
        ```
    """
    # Calculate the coagulation probabilities for each particle pair
    coagulation_probabilities = kernel_values / kernel_max
    # coagulation_probabilities = kernel_values * kernel_max

    # Determine which events occur based on these probabilities
    coagulation_occurs = (
        generator.uniform(low=0, high=1, size=len(coagulation_probabilities))
        < coagulation_probabilities
    )
    # coagulation_occurs = coagulation_probabilities>0

    # Return the indices of particles that underwent coagulation
    return small_index[coagulation_occurs], large_index[coagulation_occurs]


def _sort_particles(
    particle_radius: NDArray[np.float64],
    particle_concentration: Optional[NDArray[np.float64]] = None,
) -> Tuple[
    NDArray[np.int64], NDArray[np.float64], Optional[NDArray[np.float64]]
]:
    """Sort particle radii (and optionally concentrations) in ascending order.

    The function returns an array of `unsort_indices` that can be used
    to restore the particles to their original order after manipulations.

    Arguments:
        - particle_radius : 1D NumPy array of particle radii.
        - particle_concentration : Optional array of corresponding
          concentrations.

    Returns:
        - `unsort_indices` : Indices to revert sorting to the original order.
        - `sorted_radius` : Sorted array of radii in ascending order.
        - `sorted_concentration` : Sorted array of concentrations, if
          provided; otherwise `None`.

    Examples:
        ```py
        import numpy as np
        r = np.array([0.3, 0.1, 0.5])
        c = np.array([10, 30, 20])
        u_idx, s_r, s_c = _sort_particles(r, c)
        # s_r -> [0.1, 0.3, 0.5]
        # s_c -> [30, 10, 20]
        # u_idx can be used to get them back in [0.3, 0.1, 0.5] order
        ```
    """
    # Sort the particle radii and get the sorted indices
    sorted_indices = np.argsort(particle_radius)
    # Calculate the indices needed to revert the sorting
    unsort_indices = np.argsort(sorted_indices)
    # Sort the particle radii
    sorted_radius = particle_radius[sorted_indices]

    # If concentrations are provided, sort them too; otherwise, return None
    sorted_concentration = (
        particle_concentration[sorted_indices]
        if particle_concentration is not None
        else None
    )
    return unsort_indices, sorted_radius, sorted_concentration


def _bin_particles(
    particle_radius: NDArray[np.float64],
    radius_bins: NDArray[np.float64],
) -> Tuple[NDArray[np.int64], NDArray[np.int64]]:
    """Divide the sorted particle radii into bins and count how many fall into
    each bin.

    This function uses `radius_bins` as edges and assigns each particle
    radius to a bin index via `np.digitize`. The result is (1) a histogram
    with the number of particles in each bin, and (2) an array of per-particle
    bin indices.

    Arguments:
        - particle_radius : Array of sorted particle radii.
        - radius_bins : Edges used to define the bins.

    Returns:
        - `number_in_bins` : Counts of how many radii lie in each bin.
        - `bin_indices` : The bin index assigned to each particle.

    Examples:
        ```py
        import numpy as np
        rad = np.array([1e-9, 1.5e-9, 2e-9, 5e-9])
        bin_edges = np.array([1e-9, 2e-9, 3e-9, 1e-8])
        n_in_bins, bin_idx = _bin_particles(rad, bin_edges)
        # n_in_bins -> [1, 2, 1]
        # bin_idx might be [0, 1, 1, 2]
        ```
    """
    number_in_bins, bins = np.histogram(particle_radius, bins=radius_bins)
    bin_indices = np.digitize(particle_radius, bins, right=True)

    # Handle overflow and underflow bin indices
    bin_indices[bin_indices == len(bins)] = len(bins) - 1
    bin_indices[bin_indices == 0] = 1
    bin_indices -= 1  # Adjust to 0-based indexing

    return number_in_bins, bin_indices


def _get_bin_pairs(
    bin_indices: NDArray[np.int64],
) -> list[Tuple[int, int]]:
    """Produce the list of all unique (binA, binB) pairs using combinations
    with replacement.

    This function is useful when we want to iterate over all bin pairs
    (including binA == binB) for collision computations. The combination
    ensures each pair is returned only once.

    Arguments:
        - bin_indices : Array of bin indices for each particle (though
          only the unique values matter).

    Returns:
        - A list of (lower_bin, upper_bin) pairs covering all unique
          bins in `bin_indices`.

    Examples:
        ```py
        bins = np.array([0, 0, 1, 2, 2])
        pairs = _get_bin_pairs(bins)
        # pairs -> [(0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2)]
        ```
    """
    unique_bins = np.unique(bin_indices)
    return list(combinations_with_replacement(unique_bins, 2))


def _calculate_concentration_in_bins(
    bin_indices: NDArray[np.int64],
    particle_concentration: NDArray[np.float64],
    number_in_bins: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Sum the particle concentrations in each bin.

    Given per-particle `bin_indices` and `particle_concentration`, this
    function accumulates the total concentration of all particles that
    fall into each bin. The `number_in_bins` array is used mainly for
    shape reference but can also confirm the count of particles.

    Arguments:
        - bin_indices : Array of bin indices for each particle.
        - particle_concentration : 1D array of concentrations matching
          each particle.
        - number_in_bins : Array with the count of particles in each bin.

    Returns:
        - A 1D array whose length is the number of unique bins, containing
          the summed concentration per bin.

    Examples:
        ```py
        import numpy as np
        b_idx = np.array([0, 0, 1, 1, 2])
        conc = np.array([10., 5., 2., 3., 4.])
        n_in_bins = np.array([2, 2, 1])  # might match the bin partition
        bin_c = _calculate_concentration_in_bins(b_idx, conc, n_in_bins)
        # bin_c -> [15., 5., 4.]
        ```
    """
    concentration_in_bins = np.zeros_like(number_in_bins, dtype=np.float64)
    unique_bins = np.unique(bin_indices)

    for unique_bin in unique_bins:
        concentration_in_bins[unique_bin] = np.sum(
            particle_concentration[bin_indices == unique_bin]
        )

    return concentration_in_bins


# pylint: disable=too-many-positional-arguments, too-many-arguments, too-many-locals
def get_super_droplet_coagulation_step(
    particle_radius: NDArray[np.float64],
    particle_concentration: NDArray[np.float64],
    kernel: NDArray[np.float64],
    kernel_radius: NDArray[np.float64],
    volume: float,
    time_step: float,
    random_generator: np.random.Generator,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Carry out one time-step of super-droplet-based coagulation.

    This function sorts particles by radius, bins them, and then stochastically
    computes collision events according to the coagulation kernel. It updates
    the particle radii/concentrations, then unsorts them back to the original
    order.

    Arguments:
        - particle_radius : Array of particle radii (m).
        - particle_concentration : Array of per-particle concentration.
        - kernel : 2D matrix of coagulation kernel values, dimension
          ~ len(kernel_radius) × len(kernel_radius).
        - kernel_radius : Array of radius points defining the kernel dimension.
        - volume : System volume or domain size in m³.
        - time_step : The length of this coagulation iteration in seconds.
        - random_generator : Random number generator for sampling collisions.

    Returns:
        - Updated radii array after processing coagulation.
        - Updated concentrations array after processing coagulation.

    Examples:
        ```py
        import numpy as np
        from numpy.random import default_rng
        radius = np.array([1e-9, 2e-9, 5e-9])
        conc = np.array([100., 50., 10.])
        ker_vals = np.ones((3,3))
        ker_r = np.array([1e-9, 2e-9, 5e-9])
        rng = default_rng(42)
        r_new, c_new = get_super_droplet_coagulation_step(
            radius, conc, ker_vals, ker_r, 1e-3, 1.0, rng)
        # r_new, c_new have updated values after one super droplet
        # coagulation step.

    References:
        - E. W. Tedford and L. A. Perugini, "Superdroplet method
          in cloud microphysics simulations," J. Atmos. Sci., 2020.
        - Seinfeld, J. H., & Pandis, S. N. *Atmospheric Chemistry and Physics*,
          Wiley, 2016.
    """
    # Step 1: Sort particles by size and obtain indices to revert sorting later
    unsort_indices, sorted_radius, sorted_concentration = _sort_particles(
        particle_radius=particle_radius,
        particle_concentration=particle_concentration,
    )
    # Step 2: Bin particles by size using the provided kernel radius bins
    number_in_bins, bin_indices = _bin_particles(
        particle_radius=sorted_radius, radius_bins=kernel_radius
    )
    # Step 3: Precompute unique bin pairs for efficient coagulation
    # computations
    pair_indices = _get_bin_pairs(bin_indices=bin_indices)

    # Step 4: Calculate the total concentration of particles within each bin
    concentration_in_bins = _calculate_concentration_in_bins(
        bin_indices=bin_indices,
        particle_concentration=sorted_concentration,  # type: ignore
        number_in_bins=number_in_bins,
    )

    # Step 5: Initialize a bivariate spline for interpolating kernel values
    # between bin radii
    interp_kernel = RectBivariateSpline(
        x=kernel_radius, y=kernel_radius, z=kernel
    )

    # Step 6: Initialize a counter to track the number of coagulation events
    # per particle
    single_event_counter = np.zeros_like(particle_radius, dtype=int)

    # Step 7: Iterate over each unique pair of bins to perform
    # coagulation events
    for lower_bin, upper_bin in pair_indices:
        # Step 7.1: Retrieve the maximum kernel value for the current bin pair
        # Note: The '+1' indexing assumes that 'kernel' has dimensions
        # accommodating this offset due to bin edges
        kernel_max = kernel[lower_bin, upper_bin + 1]

        # Step 7.2: Determine potential coagulation events between
        # particles in these bins
        events = _event_pairs(
            lower_bin=lower_bin,
            upper_bin=upper_bin,
            kernel_max=kernel_max,
            number_in_bins=concentration_in_bins,
        )

        # Step 7.3: Sample the number of coagulation events from a
        # Poisson distribution
        num_events = _sample_events(
            events=events,
            volume=volume,
            time_step=time_step,
            generator=random_generator,
        )

        # Step 7.4: If no events are expected, skip to the next bin pair
        if num_events == 0:
            continue

        # Step 7.5: Limit the number of events to the available number of
        # particles in each bin. This prevents oversampling beyond the
        # available particles
        num_events = min(
            num_events,
            number_in_bins[lower_bin],
            number_in_bins[upper_bin],
        )

        # Step 7.6: Randomly select indices of particles involved in the
        # coagulation events within the current bins
        r_i_indices, r_j_indices = _select_random_indices(
            lower_bin=lower_bin,
            upper_bin=upper_bin,
            events=num_events,
            number_in_bins=number_in_bins,
            generator=random_generator,
        )

        # Step 7.7: Convert bin-relative indices to actual particle indices
        # in the sorted arrays
        indices_i, indices_j = _bin_to_particle_indices(
            lower_indices=r_i_indices,
            upper_indices=r_j_indices,
            lower_bin=lower_bin,
            upper_bin=upper_bin,
            bin_indices=bin_indices,
        )

        # Step 7.8: Filter out invalid particle pairs based on their radii
        # and event counters
        indices_i, indices_j = _filter_valid_indices(
            small_index=indices_i,
            large_index=indices_j,
            particle_radius=particle_radius,
            single_event_counter=single_event_counter,
        )

        # Step 7.9: If no valid indices remain after filtering, skip to
        # the next bin pair
        if indices_i.size == 0:
            continue

        # Step 7.10: Interpolate kernel values for the selected particle pairs
        kernel_values = interp_kernel.ev(
            particle_radius[indices_i], particle_radius[indices_j]
        )

        # Step 7.11: Determine which coagulation events actually occur based
        # on interpolated kernel probabilities
        indices_i, indices_j = _coagulation_events(
            small_index=indices_i,
            large_index=indices_j,
            kernel_values=kernel_values,
            kernel_max=kernel_max,
            generator=random_generator,
        )

        # Step 7.12: Update particle properties based on the coagulation events
        # This step typically involves merging particles, updating
        # concentrations, and tracking events
        particle_radius, particle_concentration, single_event_counter = (
            _super_droplet_update_step(
                particle_radius=particle_radius,
                concentration=particle_concentration,
                single_event_counter=single_event_counter,
                small_index=indices_i,
                large_index=indices_j,
            )
        )

    # Step 8: Unsort the particles to restore their original ordering
    # before sorting
    particle_radius = particle_radius[unsort_indices]
    particle_concentration = particle_concentration[unsort_indices]

    # Step 9: Return the updated particle radii and concentrations after
    # the coagulation step
    return particle_radius, particle_concentration



"""Test the kappa convert module."""

import numpy as np

from particula.particles.properties.convert_kappa_volumes import (
    get_kappa_from_volumes,
    get_solute_volume_from_kappa,
    get_water_volume_from_kappa,
)


def test_kappa_volume_solute():
    """Test the kappa_volume_solute function."""
    # Test with water_activity = 0.95 and kappa = 0.4
    assert np.allclose(
        get_solute_volume_from_kappa(100, 0.4, 0.95), 11.6279, rtol=1e-4
    )

    # Test with water_activity = 1 and kappa = 0.0 (zero kappa correction)
    assert np.allclose(
        get_solute_volume_from_kappa(200, 0.0, 1), 0.0, rtol=1e-4
    )


def test_kappa_volume_water():
    """Test the kappa_volume_water function."""
    # Test with water_activity = 0.95 and kappa = 0.4
    assert np.allclose(
        get_water_volume_from_kappa(5, 0.5, 0.99), 247.49, rtol=1e-4
    )


def test_kappa_volume_solute_zero_correction():
    """Test the kappa_volume_solute_zero_correction function."""
    # Test with water_activity = 1 and kappa = 0.0 (zero correction)
    assert np.allclose(
        get_solute_volume_from_kappa(5, 0.5, 1), 1.1258e16, rtol=1e14
    )


def test_kappa_from_volume():
    """Test the kappa_from_volume function."""
    # Test with water_activity = 0.95 and kappa = 0.4
    assert np.allclose(get_kappa_from_volumes(100, 200, 0.5), 2.0, rtol=1e-2)

    # Test with water_activity = 1 (zero kappa correction)
    assert np.allclose(
        get_kappa_from_volumes(100, 200, 1), 4.44089e-16, rtol=1e-17
    )



"""Integration test for quick start example in particula."""

import numpy as np

import particula as par


def test_quick_start_runs():
    """Test the quick start example runs without errors."""
    # 1. Build the GasSpecies for an organic vapor:
    organic = (
        par.gas.GasSpeciesBuilder()
        .set_name("organic")
        .set_molar_mass(180e-3, "kg/mol")
        .set_vapor_pressure_strategy(
            par.gas.ConstantVaporPressureStrategy(1e2)  # Pa
        )
        .set_partitioning(True)
        .set_concentration(np.array([1e2]), "kg/m^3")
        .build()
    )

    # 2. Use AtmosphereBuilder to configure temperature, pressure, and species:
    atmosphere = (
        par.gas.AtmosphereBuilder()
        .set_temperature(298.15, "K")
        .set_pressure(101325, "Pa")
        .set_more_partitioning_species(organic)
        .build()
    )

    # 3. Build the particle distribution:
    #    Using PresetParticleRadiusBuilder, we set mode radius, GSD, etc.
    particle = (
        par.particles.PresetParticleRadiusBuilder()
        .set_mode(np.array([100e-9]), "m")
        .set_geometric_standard_deviation(np.array([1.2]))
        .set_number_concentration(np.array([1e8]), "1/m^3")
        .set_density(1e3, "kg/m^3")
        .build()
    )

    # 4. Create the Aerosol combining the atmosphere and particle distribution:
    aerosol = (
        par.AerosolBuilder()
        .set_atmosphere(atmosphere)
        .set_particles(particle)
        .build()
    )

    # 5. Define the isothermal condensation strategy:
    condensation_strategy = par.dynamics.CondensationIsothermal(
        molar_mass=180e-3,  # kg/mol
        diffusion_coefficient=2e-5,  # m^2/s
        accommodation_coefficient=1.0,
    )

    # 6. Build the MassCondensation process:
    process = par.dynamics.MassCondensation(condensation_strategy)

    # 7. Execute the condensation process over 10 seconds:
    result = process.execute(aerosol, time_step=10.0)

    #   The result is an Aerosol instance with updated particle properties.
    # print(result)
    # 8. Confirm the result is an Aerosol instance:
    assert isinstance(result, par.Aerosol)



"""Module to calculate the Kelvin effect on vapor pressure."""

from typing import Union

import numpy as np
from numpy.typing import NDArray

from particula.util.constants import GAS_CONSTANT
from particula.util.machine_limit import get_safe_exp
from particula.util.validate_inputs import validate_inputs

# Maximum Kelvin ratio to prevent overflow in exponential calculation
# exp(100) ≈ 2.7e43 is extremely large but numerically stable
# For particles smaller than ~0.1 nm, continuum mechanics breaks down
MAX_KELVIN_RATIO = 100.0


@validate_inputs(
    {
        "effective_surface_tension": "positive",
        "effective_density": "positive",
        "molar_mass": "positive",
        "temperature": "positive",
    }
)
def get_kelvin_radius(
    effective_surface_tension: Union[float, NDArray[np.float64]],
    effective_density: Union[float, NDArray[np.float64]],
    molar_mass: Union[float, NDArray[np.float64]],
    temperature: float,
) -> Union[float, NDArray[np.float64]]:
    """Compute the Kelvin radius (rₖ) to account for curvature effects on vapor
    pressure.

    The Kelvin radius is defined by:

    - rₖ = (2 × σ × M) / (R × T × ρ)
        - rₖ is Kelvin radius in meters (m).
        - σ is the effective surface tension in N/m.
        - M is the molar mass in kg/mol.
        - R is the universal gas constant in J/(mol·K).
        - T is the temperature in Kelvin (K).
        - ρ is the effective density in kg/m³.

    Arguments:
        - effective_surface_tension : Surface tension of the mixture (N/m).
        - effective_density : Effective density of the mixture (kg/m³).
        - molar_mass : Molar mass (kg/mol).
        - temperature : Temperature of the system (K).

    Returns:
        - Kelvin radius in meters (float or NDArray[np.float64]).

    Examples:
        ``` py title="Example"
        import numpy as np
        import particula as par
        par.particles.get_kelvin_radius(
            effective_surface_tension=0.072,
            effective_density=1000.0,
            molar_mass=0.018,
            temperature=298.15
        )
        # Output: ...
        ```

    References:
        - "Kelvin equation," Wikipedia,
          https://en.wikipedia.org/wiki/Kelvin_equation
    """
    return (2 * effective_surface_tension * molar_mass) / (
        GAS_CONSTANT * temperature * effective_density
    )


@validate_inputs(
    {
        "particle_radius": "nonnegative",
        "kelvin_radius_value": "nonnegative",
    }
)
def get_kelvin_term(
    particle_radius: Union[float, NDArray[np.float64]],
    kelvin_radius_value: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute the Kelvin exponential term to account for curvature effects.

    The Kelvin term (K) is given by:

    - K = exp(rₖ / rₚ)
        - K is dimensionless.
        - rₖ is the Kelvin radius (m).
        - rₚ is the particle radius (m).

    For very small particles (< 1 nm), the ratio rₖ / rₚ can become
    extremely large, leading to numerical overflow. To prevent this, the
    ratio is clipped to a maximum value of 100, corresponding to a Kelvin
    term of ~2.7e43. This is physically unrealistic but ensures numerical
    stability. Below ~0.1 nm, continuum mechanics breaks down anyway, so
    the condensation equations become questionable.

    Arguments:
        - particle_radius : Radius of the particle (m).
        - kelvin_radius_value : Precomputed Kelvin radius (m).

    Returns:
        - Dimensionless exponential factor adjusting vapor pressure.
          For extreme cases, the value is clipped to prevent overflow.

    Examples:
        ``` py title="Example"
        import particula as par
        par.particles.get_kelvin_term(
            particle_radius=1e-7,
            kelvin_radius_value=2e-7
        )
        print(kv_term)
        # Output: ...
        ```

    References:
        - Donahue, N. M., et al. (2013). "How do organic vapors contribute to
          new-particle formation?" Faraday Discussions, 165, 91–104.
          https://doi.org/10.1039/C3FD00046J. [check]
    """
    kelvin_expand = False
    # Broadcast the arrays if necessary np.isscalar(kelvin_radius_value)
    if isinstance(kelvin_radius_value, np.ndarray) and (
        kelvin_radius_value.size > 1
    ):
        kelvin_expand = True
        kelvin_radius_value = kelvin_radius_value[np.newaxis, :]

    # Suppress divide-by-zero warnings - zero radius is handled by clipping
    with np.errstate(divide="ignore", invalid="ignore"):
        if isinstance(particle_radius, np.ndarray) and not kelvin_expand:
            kelvin_ratio = kelvin_radius_value / particle_radius
            kelvin_ratio = np.clip(kelvin_ratio, None, MAX_KELVIN_RATIO)
            return get_safe_exp(kelvin_ratio)
        if (
            isinstance(particle_radius, np.ndarray)
            and (particle_radius.size > 1)
            and kelvin_expand
        ):
            particle_radius = particle_radius[:, np.newaxis]
            kelvin_ratio = kelvin_radius_value / particle_radius
            kelvin_ratio = np.clip(kelvin_ratio, None, MAX_KELVIN_RATIO)
            return get_safe_exp(kelvin_ratio)

        # Scalar case
        kelvin_ratio = kelvin_radius_value / particle_radius
        kelvin_ratio = np.clip(kelvin_ratio, None, MAX_KELVIN_RATIO)
        return get_safe_exp(kelvin_ratio)



"""
Import all the utility classes and functions, so they can be accessed from
'from particula import util'
"""

# pylint: disable=unused-import
# flake8: noqa
# pyright: basic

from particula.util.arbitrary_round import (
    get_arbitrary_round,
)
from particula.util import constants
from particula.util.machine_limit import (
    get_safe_exp,
    get_safe_log,
    get_safe_log10,
)
from particula.util.reduced_quantity import (
    get_reduced_self_broadcast,
    get_reduced_value,
)
from particula.util.refractive_index_mixing import (
    get_effective_refractive_index,
)
from particula.util.convert_dtypes import (
    get_coerced_type,
    get_dict_from_list,
    get_shape_check,
    get_values_of_dict,
)
from particula.util.convert_units import (
    get_unit_conversion,
)
from particula.util.colors import (
    TAILWIND,
)
from particula.util.chemical import (
    get_chemical_search,
    get_chemical_surface_tension,
    get_chemical_vapor_pressure,
    get_chemical_stp_properties,
)



"""Velocity correlation terms for the DNS kernel of the turbulent coagulation
model by Ayala 2008.

- Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
    the geometric collision rate of sedimenting droplets. Part 2.
    Theory and parameterization. New Journal of Physics, 10.
    https://doi.org/10.1088/1367-2630/10/7/075016
"""

from typing import Union

import numpy as np
from numpy.typing import NDArray

from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {
        "lagrangian_taylor_microscale_time": "positive",
        "lagrangian_integral_scale": "positive",
    }
)
def compute_z(
    lagrangian_taylor_microscale_time: Union[float, NDArray[np.float64]],
    lagrangian_integral_scale: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute z, the ratio of Taylor microscale time to Lagrangian timescale.

    Where the equation is
    - z = τ_T / T_L
        - τ_T (lagrangian_taylor_microscale_time) is the Lagrangian Taylor
            microscale time [s].
        - T_L (lagrangian_integral_scale) is the Lagrangian integral
            timescale [s].

    Arguments:
        - lagrangian_taylor_microscale_time : Lagrangian Taylor microscale
            time [s].
        - lagrangian_integral_scale : Lagrangian integral timescale [s].

    Returns:
        - z value [dimensionless].

    Examples:
        ```py
        example_z = compute_z(0.5, 1.0)
        # Output: 0.5
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return lagrangian_taylor_microscale_time / lagrangian_integral_scale


@validate_inputs(
    {
        "taylor_microscale": "positive",
        "eulerian_integral_length": "positive",
    }
)
def compute_beta(
    taylor_microscale: Union[float, NDArray[np.float64]],
    eulerian_integral_length: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute β, the ratio of microscale to integral length scale.

    - β = (√2 × λ) / L_e
        - λ is Taylor microscale [m].
        - L_e is Eulerian integral length scale [m].

    Arguments:
        - taylor_microscale : Taylor microscale [m].
        - eulerian_integral_length : Eulerian integral length scale [m].

    Returns:
        - β value [dimensionless].

    Examples:
        ```py
        beta_val = compute_beta(0.001, 0.1)
        # Output: 0.01414
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return (np.sqrt(2) * taylor_microscale) / eulerian_integral_length


@validate_inputs({"z": "positive"})
def compute_b1(
    z: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute b₁, a dimensionless parameter in the Ayala 2008 model.

    - b₁ = (1 + √(1 - 2z²)) / (2 √(1 - 2z²))
        - z is τ_T / T_L.

    Arguments:
        - z : A dimensionless parameter related to turbulence [-].

    Returns:
        - b₁ value [dimensionless].

    Examples:
        ```py
        b1_val = compute_b1(0.5)
        # Output: 0.866
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    sqrt_term = np.sqrt(1 - 2 * z**2)
    return (1 + sqrt_term) / (2 * sqrt_term)


@validate_inputs({"z": "positive"})
def compute_b2(
    z: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute b₂, a dimensionless parameter in the Ayala 2008 model.

    - b₂ = (1 - √(1 - 2z²)) / (2 √(1 - 2z²))
        - z is τ_T / T_L.

    Arguments:
        - z : A dimensionless parameter related to turbulence [-].

    Returns:
        - b₂ value [dimensionless].

    Examples:
        ```py
        b2_val = compute_b2(0.5)
        # Output: 0.134
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    sqrt_term = np.sqrt(1 - 2 * z**2)
    return (1 - sqrt_term) / (2 * sqrt_term)


@validate_inputs({"z": "positive", "lagrangian_integral_scale": "positive"})
def compute_c1(
    z: Union[float, NDArray[np.float64]],
    lagrangian_integral_scale: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute c₁, a dimensionless timescale factor in the Ayala 2008 model.

    - c₁ = ((1 + √(1 - 2z²)) × T_L) / 2
        - z is τ_T / T_L.
        - T_L is the Lagrangian integral timescale [s].

    Arguments:
        - z : A dimensionless parameter related to turbulence [-].
        - lagrangian_integral_scale : Lagrangian integral timescale [s].

    Returns:
        - c₁ value [dimensionless].

    Examples:
        ```py
        c1_val = compute_c1(0.5, 1.0)
        # Output: 0.933
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return ((1 + np.sqrt(1 - 2 * z**2)) * lagrangian_integral_scale) / 2


@validate_inputs({"z": "positive", "lagrangian_integral_scale": "positive"})
def compute_c2(
    z: Union[float, NDArray[np.float64]],
    lagrangian_integral_scale: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute c₂, a dimensionless timescale factor in the Ayala 2008 model.

    - c₂ = ((1 - √(1 - 2z²)) × T_L) / 2
        - z is τ_T / T_L.
        - T_L is the Lagrangian integral timescale [s].

    Arguments:
        - z : A dimensionless parameter related to turbulence [-].
        - lagrangian_integral_scale : Lagrangian integral timescale [s].

    Returns:
        - c₂ value [dimensionless].

    Examples:
        ```py
        c2_val = compute_c2(0.5, 1.0)
        # Output: 0.067
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return ((1 - np.sqrt(1 - 2 * z**2)) * lagrangian_integral_scale) / 2


@validate_inputs({"beta": "positive"})
def compute_d1(
    beta: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute d₁, another dimensionless coefficient from Ayala 2008.

    - d₁ = (1 + √(1 - 2β²)) / (2 √(1 - 2β²))
        - β is defined as β = (√2 × λ) / L_e.

    Arguments:
        - beta : A dimensionless parameter related to turbulence [-].

    Returns:
        - d₁ value [dimensionless].

    Examples:
        ```py
        d1_val = compute_d1(0.5)
        # Output: 0.866
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    sqrt_term = np.sqrt(1 - 2 * beta**2)
    return (1 + sqrt_term) / (2 * sqrt_term)


@validate_inputs({"beta": "positive"})
def compute_d2(
    beta: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute d₂, another dimensionless coefficient from Ayala 2008.

    - d₂ = (1 - √(1 - 2β²)) / (2 √(1 - 2β²))
        - β is defined as β = (√2 × λ) / L_e.

    Arguments:
        - beta : A dimensionless parameter related to turbulence [-].

    Returns:
        - d₂ value [dimensionless].

    Examples:
        ```py
        d2_val = compute_d2(0.5)
        # Output: 0.134
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    sqrt_term = np.sqrt(1 - 2 * beta**2)
    return (1 - sqrt_term) / (2 * sqrt_term)


@validate_inputs({"beta": "positive", "eulerian_integral_length": "positive"})
def compute_e1(
    beta: Union[float, NDArray[np.float64]],
    eulerian_integral_length: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute e₁, which scales the integral length by a factor in Ayala 2008.

    - e₁ = ((1 + √(1 - 2β²)) × L_e) / 2
        - β is defined as β = (√2 × λ) / L_e.
        - L_e is the Eulerian integral length scale [m].

    Arguments:
        - beta : A dimensionless parameter related to turbulence [-].
        - eulerian_integral_length : Eulerian integral length scale [m].

    Returns:
        - e₁ value [dimensionless].

    Examples:
        ```py
        e1_val = compute_e1(0.5, 0.1)
        # Output: 0.0866
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return ((1 + np.sqrt(1 - 2 * beta**2)) * eulerian_integral_length) / 2


@validate_inputs({"beta": "positive", "eulerian_integral_length": "positive"})
def compute_e2(
    beta: Union[float, NDArray[np.float64]],
    eulerian_integral_length: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Compute e₂, which scales the integral length by a factor in Ayala 2008.

    - e₂ = ((1 - √(1 - 2β²)) × L_e) / 2
        - β is defined as β = (√2 × λ) / L_e.
        - L_e is the Eulerian integral length scale [m].

    Arguments:
        - beta : A dimensionless parameter related to turbulence [-].
        - eulerian_integral_length : Eulerian integral length scale [m].

    Returns:
        - e₂ value [dimensionless].

    Examples:
        ```py
        e2_val = compute_e2(0.5, 0.1)
        # Output: 0.0134
        ```

    References:
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
            the geometric collision rate of sedimenting droplets. Part 2.
            Theory and parameterization. New Journal of Physics, 10.
            https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return ((1 - np.sqrt(1 - 2 * beta**2)) * eulerian_integral_length) / 2



"""Module for calculating the partial pressure of a species in a
gas over particle phase.
"""

from typing import Union

import numpy as np
from numpy.typing import NDArray


def get_partial_pressure_delta(
    partial_pressure_gas: Union[float, NDArray[np.float64]],
    partial_pressure_particle: Union[float, NDArray[np.float64]],
    kelvin_term: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the difference in partial pressure between gas and particle
    phase, considering the Kelvin effect.

    - Δp = p_gas − (p_particle × K)
        - p_gas is the partial pressure in the gas phase,
        - p_particle is the partial pressure in the particle phase,
        - K is the Kelvin term (dimensionless).

    Arguments:
        - partial_pressure_gas : Partial pressure of the species in the gas
            phase.
        - partial_pressure_particle : Partial pressure of the species in
            the particle phase.
        - kelvin_term : Dimensionless Kelvin effect factor due to particle
            curvature.

    Returns:
        - The difference in partial pressure, as either a float or
            NDArray[np.float64].

    Examples:
        ``` py title="Example"
        import particula as par
        par.particles.get_partial_pressure_delta(
            partial_pressure_gas=1000.0,
            partial_pressure_particle=900.0,
            kelvin_term=1.01
        )
        # Output: 1000.0 - (900.0 * 1.01) = 91.0
        ```

    References:
        - [Kelvin effect, Wikipedia](https://en.wikipedia.org/wiki/Kelvin_equation)
        - [Partial pressure, Wikipedia](https://en.wikipedia.org/wiki/Partial_pressure)
    """
    return partial_pressure_gas - partial_pressure_particle * kelvin_term



"""Turbulent shear kernel for coagulation (Saffman & Turner, 1956).

This module computes the collision kernel attributed to turbulent shear
in a fluid. The formula stems from Equation 13A.2 in Seinfeld & Pandis (2016),
and was originally derived by Saffman & Turner (1956). Particles gain relative
velocities due to turbulent velocity gradients, increasing collision likelihood.

Equation:
    - K(D₁, D₂) = √(π × eₖ / (120 × ν)) × (D₁ + D₂)³
      - eₖ : Turbulent kinetic energy dissipation rate [m²/s³],
      - ν : Kinematic viscosity [m²/s],
      - D₁, D₂ : particle diameters [m].

References:
    - Saffman, P. G., & Turner, J. S. (1956). On the collision of drops
      in turbulent clouds. Journal of Fluid Mechanics, 1(1), 16-30.
      https://doi.org/10.1017/S0022112056000020
    - Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric chemistry and
      physics (3rd ed.). John Wiley & Sons. Chapter 13, Equation 13A.2.
"""

import numpy as np
from numpy.typing import NDArray

from particula.gas.properties.kinematic_viscosity import (
    get_kinematic_viscosity_via_system_state,
)


def get_turbulent_shear_kernel_st1956(
    particle_radius: NDArray[np.float64],
    turbulent_dissipation: float,
    kinematic_viscosity: float,
) -> NDArray[np.float64]:
    """Calculate the turbulent shear kernel (Equation 13A.2, Saffman & Turner,
    1956).

    This function implements the formula for collisions induced by turbulent
    shear. The turbulent dissipation rate and kinematic viscosity determine
    how rapidly eddies drive particle collisions.

    Equation:
    - K(D₁, D₂) = √(π × eₖ / (120 × ν)) × (D₁ + D₂)³
        - eₖ : Turbulent kinetic energy dissipation rate [m²/s³],
        - ν : Kinematic viscosity [m²/s],
        - D₁, D₂ : diameters of particles [m].

    Arguments:
        - particle_radius : Array of particle radii [m].
        - turbulent_dissipation : Turbulent energy dissipation rate [m²/s³].
        - kinematic_viscosity : Kinematic viscosity [m²/s].

    Returns:
        - Turbulent shear kernel matrix [m³/s], shape (n, n).

    Examples:
        ```py title="Example"
        import numpy as np

        r = np.array([1e-7, 2e-7])
        k_matrix = get_turbulent_shear_kernel_st1956(
            particle_radius=r,
            turbulent_dissipation=1e-3,
            kinematic_viscosity=1.5e-5
        )
        print(k_matrix)
        ```

    References:
        - Saffman, P. G., & Turner, J. S. (1956). On the collision of drops
          in turbulent clouds. Journal of Fluid Mechanics, 1(1), 16-30.
        - Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric chemistry and
          physics (3rd ed.). John Wiley & Sons. Chapter 13, Equation 13A.2.
    """
    diameter_sum_matrix = (
        particle_radius[:, np.newaxis] + particle_radius[np.newaxis, :]
    ) * 2  # Convert from radius to diameter sum

    return (
        np.pi * turbulent_dissipation / (120 * kinematic_viscosity)
    ) ** 0.5 * diameter_sum_matrix**3


def get_turbulent_shear_kernel_st1956_via_system_state(
    particle_radius: NDArray[np.float64],
    turbulent_dissipation: float,
    temperature: float,
    fluid_density: float,
) -> NDArray[np.float64]:
    """Calculate the turbulent shear kernel using system state data.

    This version derives the kinematic viscosity from the temperature and
    fluid density, then uses get_turbulent_shear_kernel_st1956 for the
    Saffman & Turner (1956) formula:

    Equation:
        - K(D₁, D₂) = √(π × eₖ / (120 × ν)) × (D₁ + D₂)³
          - eₖ : Turbulent dissipation rate [m²/s³],
          - ν : Kinematic viscosity [m²/s],
          - D₁, D₂ : particle diameters [m].

    Arguments:
        - particle_radius : Array of particle radii [m].
        - turbulent_dissipation : Turbulent dissipation rate [m²/s³].
        - temperature : Temperature [K].
        - fluid_density : Fluid density [kg/m³].

    Returns:
        - Turbulent shear kernel matrix [m³/s], shape (n, n).

    Examples:
        ```py title="Example"
        import numpy as np

        r = np.array([1e-7, 2e-7])
        kernel_matrix = get_turbulent_shear_kernel_st1956_via_system_state(
            particle_radius=r,
            turbulent_dissipation=1e-3,
            temperature=300,
            fluid_density=1.2
        )
        print(kernel_matrix)
        ```

    References:
        - Saffman, P. G., & Turner, J. S. (1956). On the collision of drops
          in turbulent clouds. Journal of Fluid Mechanics, 1(1), 16-30.
        - Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric chemistry and
          physics (3rd ed.). John Wiley & Sons. Chapter 13, Equation 13A.2.
    """
    kinematic_viscosity = get_kinematic_viscosity_via_system_state(
        temperature=temperature, fluid_density=fluid_density
    )
    return get_turbulent_shear_kernel_st1956(
        particle_radius=particle_radius,
        turbulent_dissipation=turbulent_dissipation,
        kinematic_viscosity=kinematic_viscosity,
    )



"""Module for the diffusive knudsen number."""

from typing import Union

import numpy as np
from numpy.typing import NDArray

from particula.particles.properties import coulomb_enhancement
from particula.util.constants import BOLTZMANN_CONSTANT
from particula.util.reduced_quantity import get_reduced_self_broadcast
from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {
        "particle_radius": "nonnegative",
        "particle_mass": "nonnegative",
        "friction_factor": "nonnegative",
    }
)
def get_diffusive_knudsen_number(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_mass: Union[float, NDArray[np.float64]],
    friction_factor: Union[float, NDArray[np.float64]],
    coulomb_potential_ratio: Union[float, NDArray[np.float64]] = 0.0,
    temperature: float = 298.15,
) -> Union[float, NDArray[np.float64]]:
    """Compute the diffusive Knudsen number for particle-particle interactions.

    The *diffusive* Knudsen number (Kn_d) differs from the standard Knudsen
    number. It represents the ratio of the mean particle persistence
    distance to the effective Coulombic interaction scale. Mathematically:

    - Kn_d = [ √(k_B × T × μ_red) / f_red ] / [ (rᵢ + rⱼ) × (Γ_c / Γ_k) ]
        - k_B is the Boltzmann constant (J/K).
        - T is the temperature (K).
        - μ_red is the reduced mass of particles (kg).
        - f_red is the reduced friction factor (dimensionless).
        - rᵢ + rⱼ is the sum of radii for the interacting particles (m).
        - Γ_c is the continuum-limit Coulomb enhancement factor(dimensionless).
        - Γ_k is the kinetic-limit Coulomb enhancement factor (dimensionless).

    Arguments:
        - particle_radius : Radius of the particle(s) in meters (m).
        - particle_mass : Mass of the particle(s) in kilograms (kg).
        - friction_factor : Friction factor(s) (dimensionless).
        - coulomb_potential_ratio : Coulomb potential ratio (dimensionless),
          zero if no charge.
        - temperature : Temperature of the system in Kelvin (K).

    Returns:
        - The diffusive Knudsen number, either a float or NDArray[np.float64].

    Examples:
        ```py title="Single Particle Example"
        import numpy as np
        import particula as par
        par.particles.get_diffusive_knudsen_number(
            particle_radius=1e-7,
            particle_mass=1e-17,
            friction_factor=0.8,
            coulomb_potential_ratio=0.3,
            temperature=300
        )
        # Output: 0.12...
        ```
        ```py title="Multiple Particles Example"
        import numpy as np
        import particula as par
        # Multiple particles example
        radius_arr = np.array([1e-7, 2e-7])
        mass_arr = np.array([1e-17, 2e-17])
        friction_arr = np.array([0.8, 1.1])
        potential_arr = np.array([0.3, 0.5])
        par.particles.par.get_diffusive_knudsen_number(
            radius_arr, mass_arr, friction_arr, potential_arr
        )
        # Output: array([...])
        ```

    References:
        - Chahl, H. S., & Gopalakrishnan, R. (2019). "High potential, near free
          molecular regime Coulombic collisions in aerosols and dusty plasmas."
          Aerosol Science and Technology, 53(8), 933-957.
          https://doi.org/10.1080/02786826.2019.1614522
        - Gopalakrishnan, R., & Hogan, C. J. (2012). "Coulomb-influenced
          collisions in aerosols and dusty plasmas." Physical Review E, 85(2).
          https://doi.org/10.1103/PhysRevE.85.026410
    """
    # Calculate the pairwise sum of radii
    if isinstance(particle_radius, np.ndarray):
        sum_of_radii = particle_radius[:, np.newaxis] + particle_radius
    else:
        sum_of_radii = 2 * particle_radius

    # Calculate reduced mass
    if isinstance(particle_mass, np.ndarray):
        reduced_mass = get_reduced_self_broadcast(particle_mass)
    else:
        reduced_mass = particle_mass

    # Calculate reduced friction factor
    if isinstance(friction_factor, np.ndarray):
        reduced_friction_factor = get_reduced_self_broadcast(friction_factor)
    else:
        reduced_friction_factor = friction_factor

    # Calculate the kinetic and continuum enhancements
    kinetic_enhance = coulomb_enhancement.get_coulomb_kinetic_limit(
        coulomb_potential_ratio
    )
    continuum_enhance = coulomb_enhancement.get_coulomb_continuum_limit(
        coulomb_potential_ratio
    )

    # Final calculation of diffusive Knudsen number
    numerator = (
        np.sqrt(temperature * BOLTZMANN_CONSTANT * reduced_mass)
        / reduced_friction_factor
    )
    denominator = sum_of_radii * continuum_enhance / kinetic_enhance

    return numerator / denominator



"""Test for convert_functional_group.py."""

from particula.activity.convert_functional_group import (
    convert_to_oh_equivalent,
)


def test_convert_to_oh_eqivalent():
    """Test for convert_to_oh_equivalent function."""
    molar_mass_ratio = 18.016 / 250
    oxygen2carbon = 0.3

    new_oxygen2carbon, new_molar_mass_ratio = convert_to_oh_equivalent(
        oxygen2carbon=oxygen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=None,
    )
    assert new_oxygen2carbon == oxygen2carbon
    assert new_molar_mass_ratio == molar_mass_ratio



"""Activity strategy factories for calculating activity and partial pressure
of species in a mixture of liquids.
"""

from typing import Union

from particula.abc_factory import StrategyFactoryABC
from particula.particles.activity_builders import (
    ActivityIdealMassBuilder,
    ActivityIdealMolarBuilder,
    ActivityKappaParameterBuilder,
)
from particula.particles.activity_strategies import (
    ActivityIdealMass,
    ActivityIdealMolar,
    ActivityKappaParameter,
)


# There is a bit of work in the constructor, but it is necessary to set the
# types of the builders and strategies correctly.
class ActivityFactory(
    StrategyFactoryABC[
        Union[
            ActivityIdealMassBuilder,
            ActivityIdealMolarBuilder,
            ActivityKappaParameterBuilder,
        ],
        Union[ActivityIdealMass, ActivityIdealMolar, ActivityKappaParameter],
    ]
):
    """Factory for creating activity strategy builders for liquid mixtures.

    This class supports various strategies (e.g., mass-ideal, molar-ideal,
    kappa-parameter) to compute activity and partial pressures of species
    based on Raoult's Law or kappa hygroscopic parameter.

    Methods:
        get_builders:
            Provides a mapping from strategy type to its corresponding builder.
        get_strategy(strategy_type, parameters):
            Validates inputs and returns a strategy instance for the specified
            strategy type (e.g., 'mass_ideal', 'molar_ideal', or
            'kappa_parameter').

    Returns:
        - ActivityStrategy: Instance configured for the chosen activity
            approach.

    Raises:
        - ValueError: If the strategy type is unknown or if required parameters
          are missing or invalid.

    Examples:
        ```py title="Factory Usage Example"
        import particula as par
        factory = par.particles.ActivityFactory()
        strategy = factory.get_strategy("mass_ideal")
        result = strategy.activity([1.0, 2.0, 3.0])
        # result -> ...
        ```

    References:
    - "Raoult's Law,"
        [Wikipedia](https://en.wikipedia.org/wiki/Raoult%27s_law).
    """

    def get_builders(self):
        """Return a mapping of strategy types to their corresponding builders.

        Returns:
            dict[str, Any]: A dictionary mapping the activity strategy type
            (e.g., 'mass_ideal', 'molar_ideal', 'kappa_parameter') to a builder
            instance.

        Examples:
            ```py title="Builders Retrieval Example"
            factory = ActivityFactory()
            builder_map = factory.get_builders()
            mass_ideal_builder = builder_map["mass_ideal"]
            # mass_ideal_builder -> ActivityIdealMassBuilder()
        """
        return {
            "mass_ideal": ActivityIdealMassBuilder(),
            "molar_ideal": ActivityIdealMolarBuilder(),
            "kappa_parameter": ActivityKappaParameterBuilder(),
            "kappa": ActivityKappaParameterBuilder(),
        }



"""Wall loss coefficient calculations for particles in chambers.

This module provides functions to calculate the wall loss coefficients
for particles in a chamber, either spherical or rectangular. These
coefficients are crucial in understanding particle lifetimes, as they
define the rates at which particles deposit onto chamber walls under
various conditions (e.g., eddy diffusivity, settling velocity, diffusion
coefficient).

References:
    - Crump, J. G., & Seinfeld, J. H. (1981). TURBULENT DEPOSITION AND
      GRAVITATIONAL SEDIMENTATION OF AN AEROSOL IN A VESSEL OF ARBITRARY
      SHAPE. In J Aerosol Sct (Vol. 12, Issue 5).
      https://doi.org/10.1016/0021-8502(81)90036-7
    - Crump, J. G., Flagan, R. C., & Seinfeld, J. H. (1982). Particle wall
      loss rates in vessels. Aerosol Science and Technology, 2(3), 303-309.
      https://doi.org/10.1080/02786828308958636
    - McMurry, P. H., & Rader, D. J. (1985). Aerosol Wall Losses in
      Electrically Charged Chambers. Aerosol Science and Technology, 4(3),
      249-268. https://doi.org/10.1080/02786828508959054
"""

from typing import Tuple, Union

import numpy as np
from numpy.typing import NDArray

from particula.particles import (
    get_debye_function,
    get_diffusion_coefficient_via_system_state,
    get_particle_settling_velocity_via_system_state,
)


def get_spherical_wall_loss_coefficient(
    wall_eddy_diffusivity: Union[float, NDArray[np.float64]],
    diffusion_coefficient: Union[float, NDArray[np.float64]],
    settling_velocity: Union[float, NDArray[np.float64]],
    chamber_radius: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the particle wall loss coefficient in a spherical chamber.

    This function computes the wall loss coefficient based on a spherical
    chamber approximation. It uses the wall eddy diffusivity, particle
    diffusion coefficient, particle settling velocity, and chamber radius.
    The calculation is:

    - k = 6 × √(Dₑ × D) / (π × R) × f + vₛ × (3 / (4 × R))
        - k is the wall loss coefficient [s⁻¹],
        - Dₑ is the wall eddy diffusivity [m²/s or effective rate],
        - D is the particle diffusion coefficient [m²/s],
        - f is the Debye function evaluation (unitless),
        - vₛ is the settling velocity [m/s],
        - R is the chamber radius [m].

    Arguments:
        - wall_eddy_diffusivity : The wall eddy diffusivity (or rate) in s⁻¹.
        - diffusion_coefficient : The diffusion coefficient of the particle
            in m²/s.
        - settling_velocity : The particle settling velocity in m/s.
        - chamber_radius : The spherical chamber radius in m.

    Returns:
        - The wall loss coefficient k, in inverse seconds
           (float or NDArray[np.float64]).

    Examples:
    ```py title="Example (float inputs)"
    from particula.dynamics.properties.wall_loss_coefficient import (
        get_spherical_wall_loss_coefficient
    )

    k_value = get_spherical_wall_loss_coefficient(
        wall_eddy_diffusivity=1e-2,
        diffusion_coefficient=5e-6,
        settling_velocity=1e-4,
        chamber_radius=0.5
    )
    print(k_value)
    # Example output: 0.0012
    ```

    References:
    - Crump, J. G., Flagan, R. C., & Seinfeld, J. H. (1982). Particle wall
      loss rates in vessels. Aerosol Science and Technology, 2(3), 303-309.
      https://doi.org/10.1080/02786828308958636
    """
    debye_variable = (
        np.pi
        * settling_velocity
        / (2 * np.sqrt(wall_eddy_diffusivity * diffusion_coefficient))
    )
    return 6 * np.sqrt(wall_eddy_diffusivity * diffusion_coefficient) / (
        np.pi * chamber_radius
    ) * get_debye_function(variable=debye_variable) + settling_velocity / (
        4 * chamber_radius / 3
    )


def get_rectangle_wall_loss_coefficient(
    wall_eddy_diffusivity: Union[float, NDArray[np.float64]],
    diffusion_coefficient: Union[float, NDArray[np.float64]],
    settling_velocity: Union[float, NDArray[np.float64]],
    chamber_dimensions: Tuple[float, float, float],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the particle wall loss coefficient in a rectangular chamber.

    This function computes the wall loss coefficient (β₀) for a rectangular
    chamber of length (L), width (W), and height (H). It uses the wall eddy
    diffusivity, particle diffusion coefficient, particle settling velocity,
    and chamber dimensions:

    - β₀ ~ (some function of wall_eddy_diffusivity, diffusion_coefficient,
    settling_velocity, and L×W×H)

    Arguments:
        - wall_eddy_diffusivity : Wall eddy diffusivity in s⁻¹.
        - diffusion_coefficient : Particle diffusion coefficient in m²/s.
        - settling_velocity : Particle settling velocity in m/s.
        - chamber_dimensions : A tuple (length, width, height) in m.

    Returns:
        - The wall loss coefficient β₀ (float or NDArray[np.float64]), in s⁻¹.

    Examples:
    ```py title="Example (float inputs)"
    from particula.dynamics.properties.wall_loss_coefficient import (
        get_rectangle_wall_loss_coefficient
    )

    beta_0 = get_rectangle_wall_loss_coefficient(
        wall_eddy_diffusivity=1e-3,
        diffusion_coefficient=1e-5,
        settling_velocity=2e-4,
        chamber_dimensions=(1.0, 0.5, 0.5)
    )
    print(beta_0)
    # Example output: 0.0009
    ```

    References:
    - Crump, J. G., & Seinfeld, J. H. (1981). TURBULENT DEPOSITION AND
      GRAVITATIONAL SEDIMENTATION OF AN AEROSOL IN A VESSEL OF ARBITRARY
      SHAPE. J Aerosol Sci, 12(5).
      https://doi.org/10.1016/0021-8502(81)90036-7
    """
    length, width, height = chamber_dimensions  # Unpack the dimensions tuple

    # Using 1/tanh(x) for coth(x)
    coth_term = 1 / np.tanh(
        (np.pi * settling_velocity)
        / (4 * np.sqrt(wall_eddy_diffusivity * diffusion_coefficient))
    )

    # Calculate the wall loss coefficient
    return (length * width * height) ** -1 * (
        4
        * height
        * (length + width)
        * np.sqrt(wall_eddy_diffusivity * diffusion_coefficient)
        / np.pi
        + settling_velocity * length * width * coth_term
    )


# pylint: disable=too-many-positional-arguments, too-many-arguments
def get_spherical_wall_loss_coefficient_via_system_state(
    wall_eddy_diffusivity: float,
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
    chamber_radius: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate spherical wall loss coefficient via system state.

    Uses the system's physical conditions (particle radius, density,
    temperature, pressure) to compute the needed diffusion and settling velocity
    before calculating the spherical wall loss coefficient:

    - k = f(
        wall_eddy_diffusivity,
        diffusion_coefficient_via_system_state,
        settling_velocity_via_system_state,
        chamber_radius
    )

    Arguments:
        - wall_eddy_diffusivity : Wall eddy diffusivity in s⁻¹.
        - particle_radius : Particle radius in m.
        - particle_density : Particle density in kg/m³.
        - temperature : System temperature in K.
        - pressure : System pressure in Pa.
        - chamber_radius : Chamber radius in m.

    Returns:
        - The computed wall loss coefficient k (float or NDArray[np.float64])
            in s⁻¹.

    Examples:
    ```py title="Example"
    from particula.dynamics.properties.wall_loss_coefficient import (
        get_spherical_wall_loss_coefficient_via_system_state
    )

    k_value = get_spherical_wall_loss_coefficient_via_system_state(
        wall_eddy_diffusivity=1e-2,
        particle_radius=1e-7,
        particle_density=1000,
        temperature=298,
        pressure=101325,
        chamber_radius=0.5
    )
    print(k_value)
    # Example output: 0.0018
    ```

    References:
    - Crump, J. G., Flagan, R. C., & Seinfeld, J. H. (1982). Particle wall
      loss rates in vessels. Aerosol Science and Technology, 2(3), 303-309.
      https://doi.org/10.1080/02786828308958636
    """
    # Step 1: Get particle diffusion coefficient
    diffusion_coefficient = get_diffusion_coefficient_via_system_state(
        particle_radius=particle_radius,
        temperature=temperature,
        pressure=pressure,
    )

    # Step 2: Get particle settling velocity
    settling_velocity = get_particle_settling_velocity_via_system_state(
        particle_radius=particle_radius,
        particle_density=particle_density,
        temperature=temperature,
        pressure=pressure,
    )

    # Step 3: Calculate the wall loss coefficient for the spherical chamber
    return get_spherical_wall_loss_coefficient(
        wall_eddy_diffusivity=wall_eddy_diffusivity,
        diffusion_coefficient=diffusion_coefficient,
        settling_velocity=settling_velocity,
        chamber_radius=chamber_radius,
    )


# pylint: disable=too-many-positional-arguments, too-many-arguments
def get_rectangle_wall_loss_coefficient_via_system_state(
    wall_eddy_diffusivity: float,
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
    chamber_dimensions: Tuple[float, float, float],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the wall loss coefficient for a rectangular chamber based on
    the system state.

    This function computes the wall loss coefficient for a rectangular chamber
    using the system's physical state, including the wall eddy diffusivity,
    particle properties (radius, density), and environmental conditions
    (temperature, pressure). The chamber dimensions (length, width, height)
    are also considered.

    Arguments:
        wall_eddy_diffusivity: The rate of wall eddy diffusivity in inverse
            seconds (s⁻¹).
        particle_radius: The radius of the particle in meters (m).
        particle_density: The density of the particle in kilograms per cubic
            meter (kg/m³).
        temperature: The temperature of the system in Kelvin (K).
        pressure: The pressure of the system in Pascals (Pa).
        chamber_dimensions: A tuple containing the length, width, and height
            of the rectangular chamber in meters (m).

    Returns:
        The calculated wall loss coefficient for the rectangular chamber.

    References:
        - Crump, J. G., & Seinfeld, J. H. (1981). TURBULENT DEPOSITION AND
            GRAVITATIONAL SEDIMENTATION OF AN AEROSOL IN A VESSEL OF ARBITRARY
            SHAPE. In J Aerosol Sct (Vol. 12, Issue 5).
            https://doi.org/10.1016/0021-8502(81)90036-7
    """
    # Step 1: Get particle settling velocity
    settling_velocity = get_particle_settling_velocity_via_system_state(
        particle_radius=particle_radius,
        particle_density=particle_density,
        temperature=temperature,
        pressure=pressure,
    )

    # Step 2: Get particle diffusion coefficient
    diffusion_coefficient = get_diffusion_coefficient_via_system_state(
        particle_radius=particle_radius,
        temperature=temperature,
        pressure=pressure,
    )

    # Step 3: Calculate and return the wall loss coefficient for the chamber
    return get_rectangle_wall_loss_coefficient(
        wall_eddy_diffusivity=wall_eddy_diffusivity,
        diffusion_coefficient=diffusion_coefficient,
        settling_velocity=settling_velocity,
        chamber_dimensions=chamber_dimensions,
    )



"""AerosolBuilder Module.

Provides a fluent interface to create an `Aerosol` instance from an
`Atmosphere` and a `ParticleRepresentation`.  The builder validates
consistency between the gas and particle phases (e.g., the number of
partitioning species) before constructing the final object.

Examples:
```py title="Typical Usage"
from particula.aerosol_builder import AerosolBuilder
aerosol = (
    AerosolBuilder()
    .set_atmosphere(my_atmosphere)
    .set_particles(my_particles)
    .build()
)
```
"""

import logging
from typing import Optional

from particula.abc_builder import BuilderABC
from particula.aerosol import Aerosol
from particula.gas.atmosphere import Atmosphere
from particula.particles.representation import ParticleRepresentation

logger = logging.getLogger("particula")


class AerosolBuilder(BuilderABC):
    """Fluent builder for `Aerosol` objects.

    The builder collects the required components—`Atmosphere` and
    `ParticleRepresentation`—and validates that they are mutually
    consistent before instantiating an `Aerosol`.

    Attributes:
        - atmosphere : Working copy of the atmosphere to embed in the aerosol.
        - particles : Working copy of the particle representation.

    Methods:
    - set_atmosphere : Assign the atmosphere component.
    - set_particles  : Assign the particle component.
    - build          : Validate and return a fully-formed `Aerosol`.

    Examples:
        ```py title="Builder Pattern"
        builder = AerosolBuilder()
        builder.set_atmosphere(atm).set_particles(prt)
        aerosol = builder.build()
        ```
    """

    def __init__(self):
        """Initialize an empty builder.

        Returns:
        - None
        """
        required_parameters = ["atmosphere", "particles"]
        BuilderABC.__init__(self, required_parameters)
        self.atmosphere: Optional[Atmosphere] = None
        self.particles: Optional[ParticleRepresentation] = None

    def set_atmosphere(
        self, atmosphere: Atmosphere, atmosphere_units: Optional[str] = None
    ) -> "AerosolBuilder":
        """Attach an `Atmosphere` to the builder.

        Args:
            atmosphere: Atmosphere to embed in the aerosol.
            atmosphere_units: Ignored; included for signature uniformity.

        Returns:
            AerosolBuilder: The builder instance (for chaining).

        Examples:
            ```py
            builder = AerosolBuilder().set_atmosphere(atm)
            ```
        """
        if atmosphere_units is not None:
            logger.warning("Ignoring units for atmosphere parameter.")
        self.atmosphere = atmosphere
        return self

    def set_particles(
        self,
        particles: ParticleRepresentation,
        particles_units: Optional[str] = None,
    ) -> "AerosolBuilder":
        """Attach a `ParticleRepresentation` to the builder.

        Args:
            particles: Particle representation to embed in the aerosol.
            particles_units: Ignored; included for signature uniformity.

        Returns:
            The builder instance (for chaining).
        """
        if particles_units is not None:
            logger.warning("Ignoring units for particles parameter.")
        self.particles = particles
        return self

    def _validate_species_length(self) -> None:
        """Validate species-count consistency between atmosphere and particles.

        When the particle distribution strategy is either
        `SpeciatedMassMovingBin` or `ParticleResolvedSpeciatedMass`, the number
        of partitioning gas species must equal the number of particle-phase
        species.

        Raises:
            - ValueError : If the species counts are inconsistent.
        """
        if self.atmosphere is None or self.particles is None:
            raise ValueError(
                "Atmosphere and particles must be set before validation."
            )
        strategy_name = self.particles.get_strategy_name()
        if strategy_name not in (
            "SpeciatedMassMovingBin",
            "ParticleResolvedSpeciatedMass",
        ):
            return  # no check required for other strategies

        partitioning_species = self.atmosphere.partitioning_species
        if partitioning_species is None:
            return  # nothing to validate

        # count partitioning species in the atmosphere
        n_partitioning = len(partitioning_species)

        # count species represented by the particle model
        species_mass = self.particles.get_species_mass(clone=False)
        n_particle_species = (
            1 if species_mass.ndim == 1 else species_mass.shape[1]
        )

        if n_partitioning != n_particle_species:
            msg = (
                "Number of partitioning species in atmosphere "
                f"({n_partitioning}) must match number of species in "
                f"particle representation ({n_particle_species}) when using "
                f"{strategy_name}."
            )
            logger.error(msg)
            raise ValueError(msg)

    def build(self) -> Aerosol:
        """Finalize and construct an `Aerosol` object.

        Raises:
            - ValueError : If either component is missing or validation fails.

        Returns:
            Aerosol object, The validated aerosol instance.

        Examples:
        ```py
        aerosol = (
            AerosolBuilder()
            .set_atmosphere(atm)
            .set_particles(prt)
            .build()
        )
        ```
        """
        self.pre_build_check()
        self._validate_species_length()
        return Aerosol(atmosphere=self.atmosphere, particles=self.particles)  # type: ignore



"""Import all the gas classes and functions, so they can be accessed from
particula.next import gas
"""

# pylint: disable=unused-import
# flake8: noqa
# pyright: basic

from particula.gas.vapor_pressure_strategies import (
    ConstantVaporPressureStrategy,
    AntoineVaporPressureStrategy,
    ClausiusClapeyronStrategy,
    WaterBuckStrategy,
    ArblasterLiquidVaporPressureStrategy,
    LiquidClausiusHybridStrategy,
    TableVaporPressureStrategy,
)

from particula.gas.vapor_pressure_builders import (
    ConstantVaporPressureBuilder,
    AntoineVaporPressureBuilder,
    ClausiusClapeyronVaporPressureBuilder,
    WaterBuckVaporPressureBuilder,
    SaturationConcentrationVaporPressureBuilder,
    TableVaporPressureBuilder,
)
from particula.gas.vapor_pressure_factories import VaporPressureFactory
from particula.gas.species import GasSpecies

from particula.gas.species_builders import (
    GasSpeciesBuilder,
    PresetGasSpeciesBuilder,
)
from particula.gas.species_factories import GasSpeciesFactory
from particula.gas.atmosphere import Atmosphere

from particula.gas.atmosphere_builders import AtmosphereBuilder

# particla.gas.properties
from particula.gas.properties.concentration_function import (
    get_concentration_from_pressure,
)
from particula.gas.properties.dynamic_viscosity import (
    get_dynamic_viscosity,
)
from particula.gas.properties.fluid_rms_velocity import get_fluid_rms_velocity
from particula.gas.properties.integral_scale_module import (
    get_eulerian_integral_length,
    get_lagrangian_integral_time,
)
from particula.gas.properties.kinematic_viscosity import (
    get_kinematic_viscosity,
    get_kinematic_viscosity_via_system_state,
)
from particula.gas.properties.kolmogorov_module import (
    get_kolmogorov_length,
    get_kolmogorov_time,
    get_kolmogorov_velocity,
)
from particula.gas.properties.mean_free_path import (
    get_molecule_mean_free_path,
)
from particula.gas.properties.normalize_accel_variance import (
    get_normalized_accel_variance_ao2008,
)
from particula.gas.properties.pressure_function import (
    get_partial_pressure,
    get_saturation_ratio_from_pressure,
)
from particula.gas.properties.taylor_microscale_module import (
    get_lagrangian_taylor_microscale_time,
    get_taylor_microscale,
    get_taylor_microscale_reynolds_number,
)
from particula.gas.properties.thermal_conductivity import (
    get_thermal_conductivity,
)
from particula.gas.properties.vapor_pressure_module import (
    get_antoine_vapor_pressure,
    get_buck_vapor_pressure,
    get_clausius_clapeyron_vapor_pressure,
)



"""Tests for the :class:`MassBasedMovingBin` distribution strategy."""

# pylint: disable=R0801

import numpy as np
import pytest

from particula.particles.distribution_strategies import MassBasedMovingBin

mass_based_strategy = MassBasedMovingBin()


def test_get_name():
    """Test retrieving the class name."""
    assert mass_based_strategy.get_name() == "MassBasedMovingBin"


def test_get_mass():
    """Test that mass equals distribution."""
    distribution = np.array([100, 200, 300], dtype=np.float64)
    density = np.array([1, 2, 3], dtype=np.float64)
    expected_mass = distribution
    np.testing.assert_array_equal(
        mass_based_strategy.get_mass(distribution, density), expected_mass
    )


def test_get_radius():
    """Test radius calculation."""
    distribution = np.array([100, 200, 300], dtype=np.float64)
    density = np.array([1, 2, 3], dtype=np.float64)
    expected = (3 * distribution / (4 * np.pi * density)) ** (1 / 3)
    np.testing.assert_allclose(
        mass_based_strategy.get_radius(distribution, density), expected
    )


def test_get_total_mass():
    """Test total mass computation."""
    distribution = np.array([100, 200, 300], dtype=np.float64)
    concentration = np.array([1, 2, 3], dtype=np.float64)
    density = np.array([1, 2, 3], dtype=np.float64)
    expected_total_mass = np.sum(distribution * concentration)
    assert mass_based_strategy.get_total_mass(
        distribution, concentration, density
    ) == pytest.approx(expected_total_mass)


def test_add_mass():
    """Test mass addition updates distribution."""
    distribution = np.array([100, 200, 300], dtype=np.float64)
    concentration = np.array([1, 2, 3], dtype=np.float64)
    density = np.array([1, 2, 3], dtype=np.float64)
    added_mass = np.array([10, 20, 30], dtype=np.float64)
    expected_dist = distribution + added_mass
    expected_conc = concentration
    new_dist, new_conc = mass_based_strategy.add_mass(
        distribution, concentration, density, added_mass
    )
    np.testing.assert_array_equal(new_dist, expected_dist)
    np.testing.assert_array_equal(new_conc, expected_conc)


def test_add_concentration():
    """Test concentration addition."""
    distribution = np.array([1.0, 2.0], dtype=np.float64)
    concentration = np.array([10.0, 20.0], dtype=np.float64)
    added_distribution = np.array([1.0, 2.0], dtype=np.float64)
    added_concentration = np.array([5.0, 5.0], dtype=np.float64)
    expected = concentration + added_concentration
    _, new_conc = mass_based_strategy.add_concentration(
        distribution,
        concentration,
        added_distribution,
        added_concentration,
    )
    np.testing.assert_array_equal(new_conc, expected)


def test_add_concentration_distribution_error():
    """Test shape validation for distribution input."""
    distribution = np.array([1.0, 2.0], dtype=np.float64)
    concentration = np.array([1.0, 2.0], dtype=np.float64)
    with pytest.raises(ValueError):
        mass_based_strategy.add_concentration(
            distribution,
            concentration,
            np.array([[1.0, 2.0]], dtype=np.float64),
            np.array([1.0, 1.0], dtype=np.float64),
        )


def test_add_concentration_shape_error():
    """Test shape validation for concentration."""
    distribution = np.array([1.0, 2.0], dtype=np.float64)
    concentration = np.array([1.0, 2.0], dtype=np.float64)
    with pytest.raises(ValueError):
        mass_based_strategy.add_concentration(
            distribution,
            concentration,
            distribution,
            np.array([[1.0, 1.0]], dtype=np.float64),
        )


def test_collide_pairs():
    """Test collide_pairs not implemented."""
    with pytest.raises(NotImplementedError):
        mass_based_strategy.collide_pairs(
            np.array([100, 200, 300], dtype=np.float64),
            np.array([1, 2, 3], dtype=np.float64),
            np.array([100, 200, 300], dtype=np.float64),
            np.array([[0, 1]], dtype=np.int64),
        )



"""Tests for the get_particle_inertia_time function."""

import numpy as np
import pytest

from particula.particles.properties.inertia_time import (
    get_particle_inertia_time,
)


def test_get_particle_inertia_time_scalar():
    """Test get_particle_inertia_time with scalar inputs."""
    particle_radius = 50e-6  # 50 microns
    particle_density = 1000  # kg/m³ (e.g., water, dust)
    fluid_density = 1.2  # kg/m³ (air)
    kinematic_viscosity = 1.5e-5  # m²/s

    expected = (
        (2 / 9)
        * (particle_density / fluid_density)
        * (particle_radius**2 / kinematic_viscosity)
    )

    result = get_particle_inertia_time(
        particle_radius,
        particle_density,
        fluid_density,
        kinematic_viscosity,
    )

    assert np.isclose(result, expected, atol=1e-10)


def test_get_particle_inertia_time_array():
    """Test get_particle_inertia_time with NumPy array inputs."""
    particle_radius = np.array([50e-6, 80e-6])
    particle_density = np.array([1000, 1200])
    fluid_density = np.array([1.2, 1.2])
    kinematic_viscosity = np.array([1.5e-5, 1.5e-5])

    expected = (
        (2 / 9)
        * (particle_density / fluid_density)
        * (particle_radius**2 / kinematic_viscosity)
    )

    result = get_particle_inertia_time(
        particle_radius,
        particle_density,
        fluid_density,
        kinematic_viscosity,
    )

    assert np.allclose(result, expected, atol=1e-10)


def test_get_particle_inertia_time_invalid():
    """Test that get_particle_inertia_time raises errors for invalid inputs."""
    with pytest.raises(ValueError):
        get_particle_inertia_time(-50e-6, 1000, 1.2, 1.5e-5)  # Negative radius

    with pytest.raises(ValueError):
        get_particle_inertia_time(
            50e-6, -1000, 1.2, 1.5e-5
        )  # Negative particle density

    with pytest.raises(ValueError):
        get_particle_inertia_time(
            50e-6, 1000, -1.2, 1.5e-5
        )  # Negative fluid density

    with pytest.raises(ValueError):
        get_particle_inertia_time(
            50e-6, 1000, 1.2, -1.5e-5
        )  # Negative viscosity



"""Coagulation Factory Module."""

from typing import Any, Dict, Union

from particula.abc_factory import StrategyFactoryABC

from .coagulation_builder.brownian_coagulation_builder import (
    BrownianCoagulationBuilder,
)
from .coagulation_builder.charged_coagulation_builder import (
    ChargedCoagulationBuilder,
)
from .coagulation_builder.combine_coagulation_strategy_builder import (
    CombineCoagulationStrategyBuilder,
)
from .coagulation_builder.turbulent_dns_coagulation_builder import (
    TurbulentDNSCoagulationBuilder,
)
from .coagulation_builder.turbulent_shear_coagulation_builder import (
    TurbulentShearCoagulationBuilder,
)
from .coagulation_strategy.coagulation_strategy_abc import (
    CoagulationStrategyABC,
)


class CoagulationFactory(
    StrategyFactoryABC[
        Union[
            BrownianCoagulationBuilder,
            ChargedCoagulationBuilder,
            TurbulentShearCoagulationBuilder,
            TurbulentDNSCoagulationBuilder,
            CombineCoagulationStrategyBuilder,
        ],
        CoagulationStrategyABC,
    ]
):
    """Factory class for creating coagulation strategy instances
    based on a given type string. Supported types include:
        - 'brownian'
        - 'charged'
        - 'turbulent_shear'
        - 'turbulent_dns'
        - 'combine'.

    Methods:
    - get_builders() : Returns the mapping of strategy types to builder
        instances.
    - get_strategy(strategy_type, parameters): Gets the strategy instance
        for the specified strategy type.
        - strategy_type: Type of coagulation strategy to use, can be
            'brownian', 'charged', 'turbulent_shear', 'turbulent_dns', or
            'combine'.
    """

    def get_builders(self) -> Dict[str, Any]:
        """Return the mapping of strategy types to builder instances.

        Returns:
            Dictionary mapping strategy names to their corresponding builders.
        """
        return {
            "brownian": BrownianCoagulationBuilder(),
            "charged": ChargedCoagulationBuilder(),
            "turbulent_shear": TurbulentShearCoagulationBuilder(),
            "turbulent_dns": TurbulentDNSCoagulationBuilder(),
            "combine": CombineCoagulationStrategyBuilder(),
        }



"""Calculate a fluids integral scale."""

from typing import Union

import numpy as np
from numpy.typing import NDArray

from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {"fluid_rms_velocity": "positive", "turbulent_dissipation": "positive"}
)
def get_lagrangian_integral_time(
    fluid_rms_velocity: Union[float, NDArray[np.float64]],
    turbulent_dissipation: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the Lagrangian integral timescale.

    The Lagrangian integral timescale is a measure of the time it takes for
    a fluid particle to travel a distance equal to the integral length scale.

    - T_L = (u'²) / ε
        - T_L is Lagrangian integral timescale [s].
        - fluid_rms_velocity (u') is Fluid RMS fluctuation velocity [m/s].
        - turbulent_dissipation (ε) is Turbulent energy dissipation rate
            [m²/s³].

    Arguments:
        - fluid_rms_velocity : Fluid RMS fluctuation velocity [m/s].
        - turbulent_dissipation : Turbulent kinetic energy dissipation rate
            [m²/s³].

    Returns:
        - Lagrangian integral timescale [s].

    Examples:
        ``` py title="Example Usage"
        import particula as par
        par.gas.get_lagrangian_integral_time(0.3, 1e-4)
        # Output: 900.0
        ```

    References:
        - Townsend, A. A., "The Structure of Turbulent Shear Flow," 2nd ed.,
          Cambridge University Press, 1976. [Check this reference]
        - Wikipedia contributors, "Turbulence," Wikipedia.
    """
    return (fluid_rms_velocity**2) / turbulent_dissipation


@validate_inputs(
    {"fluid_rms_velocity": "positive", "turbulent_dissipation": "positive"}
)
def get_eulerian_integral_length(
    fluid_rms_velocity: Union[float, NDArray[np.float64]],
    turbulent_dissipation: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the Eulerian integral length scale.

    The Eulerian integral length scale is a measure of the size of the largest
    turbulent eddies in a fluid flow.

    - L_e = 0.5 × (u'³) / ε
        - L_e is Eulerian integral length scale [m].
        - fluid_rms_velocity (u') is Fluid RMS fluctuation velocity [m/s].
        - turbulent_dissipation (ε) is Turbulent energy dissipation rate
            [m²/s³].

    Arguments:
        - fluid_rms_velocity : Fluid RMS fluctuation velocity [m/s].
        - turbulent_dissipation : Turbulent kinetic energy dissipation rate
            [m²/s³].

    Returns:
        - Eulerian integral length scale [m].

    Examples:
        ``` py title="Example"
        import particula as par
        par.gas.get_eulerian_integral_length(0.3, 1e-4)
        # Output: 1350.0
        ```

    References:
        - Hinze, J. O., "Turbulence," McGraw-Hill, 1975. [Check this reference]
        - Wikipedia contributors, "Turbulence," Wikipedia.
    """
    return 0.5 * (fluid_rms_velocity**3) / turbulent_dissipation



"""Import all the particle classes and functions, so they can be accessed from
'from particula import particles'
"""

# pylint: disable=unused-import
# flake8: noqa
# pyright: basic

from particula.particles.distribution_strategies import (
    MassBasedMovingBin,
    RadiiBasedMovingBin,
    SpeciatedMassMovingBin,
    ParticleResolvedSpeciatedMass,
)
from particula.particles.distribution_builders import (
    MassBasedMovingBinBuilder,
    RadiiBasedMovingBinBuilder,
    SpeciatedMassMovingBinBuilder,
    ParticleResolvedSpeciatedMassBuilder,
)
from particula.particles.distribution_factories import (
    DistributionFactory,
)
from particula.particles.activity_strategies import (
    ActivityIdealMass,
    ActivityIdealMolar,
    ActivityKappaParameter,
)
from particula.particles.activity_builders import (
    ActivityIdealMassBuilder,
    ActivityIdealMolarBuilder,
    ActivityKappaParameterBuilder,
)
from particula.particles.activity_factories import (
    ActivityFactory,
)
from particula.particles.representation import (
    ParticleRepresentation,
)
from particula.particles.representation_builders import (
    ParticleMassRepresentationBuilder,
    ParticleRadiusRepresentationBuilder,
    PresetParticleRadiusBuilder,
    ResolvedParticleMassRepresentationBuilder,
    PresetResolvedParticleMassBuilder,
)
from particula.particles.representation_factories import (
    ParticleRepresentationFactory,
)
from particula.particles.surface_strategies import (
    SurfaceStrategyVolume,
    SurfaceStrategyMass,
    SurfaceStrategyMolar,
)
from particula.particles.surface_builders import (
    SurfaceStrategyVolumeBuilder,
    SurfaceStrategyMassBuilder,
    SurfaceStrategyMolarBuilder,
)
from particula.particles.surface_factories import (
    SurfaceFactory,
)

# particula.particles.properties
from particula.particles.properties.activity_module import (
    get_ideal_activity_mass,
    get_ideal_activity_molar,
    get_ideal_activity_volume,
    get_kappa_activity,
    get_surface_partial_pressure,
)
from particula.particles.properties.aerodynamic_mobility_module import (
    get_aerodynamic_mobility,
)
from particula.particles.properties.aerodynamic_size import (
    AERODYNAMIC_SHAPE_FACTOR_DICT,
    get_aerodynamic_length,
    get_aerodynamic_shape_factor,
)
from particula.particles.properties.collision_radius_module import (
    get_collision_radius_mg1988,
    get_collision_radius_sr1992,
    get_collision_radius_mzg2002,
    get_collision_radius_tt2012,
    get_collision_radius_wq2022_rg,
    get_collision_radius_wq2022_rg_df,
    get_collision_radius_wq2022_rg_df_k0,
    get_collision_radius_wq2022_rg_df_k0_a13,
)
from particula.particles.properties.coulomb_enhancement import (
    get_coulomb_enhancement_ratio,
    get_coulomb_kinetic_limit,
    get_coulomb_continuum_limit,
)
from particula.particles.properties.diffusion_coefficient import (
    get_diffusion_coefficient,
    get_diffusion_coefficient_via_system_state,
)
from particula.particles.properties.diffusive_knudsen_module import (
    get_diffusive_knudsen_number,
)
from particula.particles.properties.friction_factor_module import (
    get_friction_factor,
)
from particula.particles.properties.inertia_time import (
    get_particle_inertia_time,
)
from particula.particles.properties.kelvin_effect_module import (
    get_kelvin_radius,
    get_kelvin_term,
)
from particula.particles.properties.knudsen_number_module import (
    get_knudsen_number,
)
from particula.particles.properties.lognormal_size_distribution import (
    get_lognormal_pdf_distribution,
    get_lognormal_pmf_distribution,
    get_lognormal_sample_distribution,
)
from particula.particles.properties.mean_thermal_speed_module import (
    get_mean_thermal_speed,
)
from particula.particles.properties.mixing_state_index import (
    get_mixing_state_index,
)
from particula.particles.properties.partial_pressure_module import (
    get_partial_pressure_delta,
)
from particula.particles.properties.reynolds_number import (
    get_particle_reynolds_number,
)
from particula.particles.properties.settling_velocity import (
    get_particle_settling_velocity,
    get_particle_settling_velocity_via_inertia,
    get_particle_settling_velocity_via_system_state,
    get_particle_settling_velocity_with_drag,
)
from particula.particles.properties.slip_correction_module import (
    get_cunningham_slip_correction,
)
from particula.particles.properties.special_functions import (
    get_debye_function,
)
from particula.particles.properties.stokes_number import (
    get_stokes_number,
)
from particula.particles.properties.vapor_correction_module import (
    get_vapor_transition_correction,
)
from particula.particles.properties.convert_kappa_volumes import (
    get_kappa_from_volumes,
    get_water_volume_from_kappa,
    get_solute_volume_from_kappa,
    get_water_volume_in_mixture,
)
from particula.particles.properties.convert_mass_concentration import (
    get_volume_fraction_from_mass,
    get_mass_fraction_from_mass,
    get_mole_fraction_from_mass,
)
from particula.particles.properties.convert_mole_fraction import (
    get_mass_fractions_from_moles,
)
from particula.particles.properties.convert_size_distribution import (
    get_distribution_conversion_strategy,
    get_distribution_in_dn,
    get_pdf_distribution_in_pmf,
    SameScaleConversionStrategy,
    DNdlogDPtoPDFConversionStrategy,
    DNdlogDPtoPMFConversionStrategy,
    PMFtoPDFConversionStrategy,
)
from particula.particles.properties.organic_density_module import (
    get_organic_density_estimate,
    get_organic_density_array,
)



"""Tests for mixins in coagulation_builder_mixin.py."""

import numpy as np
import pytest

# pylint: disable=line-too-long
from particula.dynamics.coagulation.coagulation_builder.coagulation_builder_mixin import (  # noqa: E501
    BuilderDistributionTypeMixin,
    BuilderFluidDensityMixin,
    BuilderTurbulentDissipationMixin,
)


class MixinTester(
    BuilderDistributionTypeMixin,
    BuilderTurbulentDissipationMixin,
    BuilderFluidDensityMixin,
):
    """Simple class to inherit and test the mixins directly."""


def test_distribution_type_valid():
    """Test setting a valid distribution type."""
    tester = MixinTester()
    tester.set_distribution_type("discrete")
    assert tester.distribution_type == "discrete"


def test_distribution_type_invalid():
    """Test that an invalid distribution type raises ValueError."""
    tester = MixinTester()
    with pytest.raises(ValueError):
        tester.set_distribution_type("invalid_type")


def test_distribution_type_units():
    """Test that providing distribution_type_units logs a warning."""
    tester = MixinTester()
    tester.set_distribution_type(
        "discrete", distribution_type_units="dimensionless"
    )
    assert tester.distribution_type == "discrete"


def test_turbulent_dissipation_negative():
    """Test that negative turbulent dissipation raises ValueError."""
    tester = MixinTester()
    with pytest.raises(ValueError):
        tester.set_turbulent_dissipation(-1, "m^2/s^3")


def test_turbulent_dissipation_positive():
    """Test that a positive turbulent dissipation is set correctly."""
    tester = MixinTester()
    tester.set_turbulent_dissipation(1.0, "m^2/s^3")
    assert tester.turbulent_dissipation == pytest.approx(1.0)


def test_fluid_density_negative():
    """Test that negative fluid density raises ValueError."""
    tester = MixinTester()
    with pytest.raises(ValueError):
        tester.set_fluid_density(-1, "kg/m^3")


def test_fluid_density_positive():
    """Test that a valid fluid density is stored correctly."""
    tester = MixinTester()
    tester.set_fluid_density(1.2, "kg/m^3")
    assert tester.fluid_density == pytest.approx(1.2)


def test_fluid_density_array():
    """Test that an array of fluid densities is set correctly."""
    tester = MixinTester()
    density_array = np.array([1.2, 1.0, 0.9])
    tester.set_fluid_density(density_array, "kg/m^3")
    np.testing.assert_array_almost_equal(tester.fluid_density, density_array)  # type: ignore



"""Integration test for isothermal condensation on particle-resolved aerosol."""

import unittest

import numpy as np

import particula as par


class TestCondensationParticleResolved(unittest.TestCase):
    """Verify that mass is transferred from gas to particles."""

    def setUp(self):
        """Set up aerosol system with gas and particles for testing."""
        # ---------- vapor-pressure strategies ----------
        self.vp_water = par.gas.VaporPressureFactory().get_strategy(
            "water_buck"
        )
        self.vp_constant = par.gas.VaporPressureFactory().get_strategy(
            "constant", {"vapor_pressure": 1e-24, "vapor_pressure_units": "Pa"}
        )

        # ---------- gas species (partitioning + gas-only) ----------
        molar_mass_water = 18.015e-3  # kg/mol
        molar_mass_am_sulf = 132.14e-3  # kg/mol
        water_sat = self.vp_water.saturation_concentration(
            molar_mass=molar_mass_water, temperature=298.15
        )
        # 2 % supersaturation
        water_conc = water_sat * 1.02

        self.gas_partitioning = (
            par.gas.GasSpeciesBuilder()
            .set_name(["H2O", "NH4HSO4"])
            .set_molar_mass(
                np.array([molar_mass_water, molar_mass_am_sulf]), "kg/mol"
            )
            .set_vapor_pressure_strategy([self.vp_water, self.vp_constant])
            .set_concentration(np.array([water_conc, 1e-30]), "kg/m^3")
            .set_partitioning(True)
            .build()
        )

        self.gas_inert = (
            par.gas.GasSpeciesBuilder()
            .set_name("N2")
            .set_molar_mass(0.028, "kg/mol")
            .set_vapor_pressure_strategy(self.vp_constant)
            .set_concentration(0.79, "kg/m^3")
            .set_partitioning(False)
            .build()
        )
        # ---------- atmosphere ----------
        self.atmosphere = (
            par.gas.AtmosphereBuilder()
            .set_temperature(298.15, "K")
            .set_pressure(101325, "Pa")
            .set_more_partitioning_species(self.gas_partitioning)
            .set_more_gas_only_species(self.gas_inert)
            .build()
        )

        # ---------- particle distribution ----------
        density_core = 1.77e3  # kg/m^3
        radii = par.particles.get_lognormal_sample_distribution(
            mode=np.array([100e-9, 400e-9]),
            geometric_standard_deviation=np.array([1.3, 1.4]),
            number_of_particles=np.array([1, 0.5]),
            number_of_samples=2_000,  # keep integration test fast
        )
        mass_core = 4 / 3 * np.pi * radii**3 * density_core
        mass_speciation = np.column_stack([mass_core * 0, mass_core])
        densities = np.array([1.0e3, density_core])

        activity = (
            par.particles.ActivityKappaParameterBuilder()
            .set_density(densities, "kg/m^3")
            .set_kappa(np.array([0.0, 0.61]))
            .set_molar_mass(
                np.array([molar_mass_water, molar_mass_am_sulf]), "kg/mol"
            )
            .set_water_index(0)
            .build()
        )
        surface = (
            par.particles.SurfaceStrategyVolumeBuilder()
            .set_density(densities, "kg/m^3")
            .set_surface_tension(np.array([0.072, 0.092]), "N/m")
            .build()
        )
        self.aerosol = par.Aerosol(
            atmosphere=self.atmosphere,
            particles=(
                par.particles.ResolvedParticleMassRepresentationBuilder()
                .set_distribution_strategy(
                    par.particles.ParticleResolvedSpeciatedMass()
                )
                .set_activity_strategy(activity)
                .set_surface_strategy(surface)
                .set_mass(mass_speciation, "kg")
                .set_density(densities, "kg/m^3")
                .set_charge(0)
                .set_volume(1e-6, "m^3")  # 1 cm³ air parcel
                .build()
            ),
        )

        # ---------- condensation process ----------
        self.condensation = par.dynamics.MassCondensation(
            condensation_strategy=par.dynamics.CondensationIsothermal(
                molar_mass=np.array([molar_mass_water, molar_mass_am_sulf]),
                diffusion_coefficient=2e-5,
                accommodation_coefficient=1,
                update_gases=True,
            )
        )

    # ------------------------------------------------------------------
    # actual tests
    # ------------------------------------------------------------------

    def test_condensation_transfers_mass(self):
        """Particle mass should increase while gas-phase water decreases."""
        initial_particle_mass = self.aerosol.particles.get_mass_concentration()
        initial_gas_water = (
            self.aerosol.atmosphere.partitioning_species.get_concentration()[0]
        )

        aerosol = self.aerosol
        for _ in range(5):
            aerosol = self.condensation.execute(aerosol, 0.1, 1)

        final_particle_mass = aerosol.particles.get_mass_concentration()
        final_gas_water = (
            aerosol.atmosphere.partitioning_species.get_concentration()[0]
        )

        # assertions
        self.assertGreater(final_particle_mass, initial_particle_mass)
        self.assertLess(final_gas_water, initial_gas_water)
        # mass conservation (water + core species)
        total_initial = initial_particle_mass + initial_gas_water
        total_final = final_particle_mass + final_gas_water
        self.assertAlmostEqual(total_initial, total_final, delta=1e-9)



"""Builders to create vapor pressure models for gas species.

This module provides builder classes for Antoine, Clausius-Clapeyron,
constant, and WaterBuck vapor pressure strategies. Each builder follows
the same workflow:

1. Configure coefficients or parameters using dedicated methods.
2. Validate required parameters.
3. Return the corresponding vapor pressure strategy object.

References:
    - "Antoine Equation,"
      [Wikipedia](https://en.wikipedia.org/wiki/Antoine_equation)
    - "Clausius–Clapeyron Relation,"
      [Wikipedia](https://en.wikipedia.org/wiki/Clausius%E2%80%93Clapeyron_relation)
    - "Vapor Pressure,"
      [Wikipedia](https://en.wikipedia.org/wiki/Vapor_pressure)
"""

import logging
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from particula.abc_builder import BuilderABC
from particula.builder_mixin import (
    BuilderMolarMassMixin,
    BuilderTemperatureMixin,
    BuilderTemperatureTableMixin,
)
from particula.gas.properties.pressure_function import get_partial_pressure
from particula.gas.vapor_pressure_strategies import (
    AntoineVaporPressureStrategy,
    ClausiusClapeyronStrategy,
    ConstantVaporPressureStrategy,
    TableVaporPressureStrategy,
    WaterBuckStrategy,
)
from particula.util import get_unit_conversion
from particula.util.validate_inputs import validate_inputs

logger = logging.getLogger("particula")


class AntoineVaporPressureBuilder(BuilderABC):
    """Builder class for AntoineVaporPressureStrategy.

    It allows setting the coefficients 'a', 'b', and 'c' separately and then
    building the strategy object.
    Follows the general form of the Antoine equation in

        log₁₀(P) = a − b / (T − c)

    Attributes:
        - a : Coefficient "a" of the Antoine equation (dimensionless).
        - b : Coefficient "b" (in Kelvin).
        - c : Coefficient "c" (in Kelvin).

    Methods:
    - set_a : Set the coefficient "a" of the Antoine equation.
    - set_b : Set the coefficient "b".
    - set_c : Set the coefficient "c".
    - build : Validate parameters and return an AntoineVaporPressureStrategy.

    Example:
        ``` py title="AntoineVaporPressureBuilder"
        strategy = (
            AntoineVaporPressureBuilder()
            .set_a(8.07131)
            .set_b(1730.63)
            .set_c(233.426)
            .build()
        )
        ```

        ``` py title="AntoineVaporPressureBuilder with units"
        strategy = (
            AntoineVaporPressureBuilder()
            .set_a(8.07131)
            .set_b(1730.63, "K")
            .set_c(233.426, "K")
            .build()
        )
        ```

    References:
        - Equation: log10(P_mmHG) = a - b / (Temperature_K - c)
          (Reference: https://en.wikipedia.org/wiki/Antoine_equation)
        - "Vapor Pressure,"
          [Wikipedia](https://en.wikipedia.org/wiki/Vapor_pressure)
        - "Atmospheric Pressure Unit Conversions,"
          [Wikipedia](https://en.wikipedia.org/wiki/Pascal_(unit))
    """

    def __init__(self):
        """Initialize the Antoine vapor pressure builder.

        Sets up the builder with required parameters (a, b, c) for
        creating an AntoineVaporPressureStrategy.
        """
        required_parameters = ["a", "b", "c"]
        # Call the base class's __init__ method
        super().__init__(required_parameters)
        self.a = None
        self.b = None
        self.c = None

    def set_a(
        self, a: float, a_units: Optional[str] = None
    ) -> "AntoineVaporPressureBuilder":
        """Set the coefficient 'a' of the Antoine equation."""
        if a < 0:
            logger.error("Coefficient 'a' must be a positive value.")
            raise ValueError("Coefficient 'a' must be a positive value.")
        if a_units is not None:
            logger.warning("Ignoring units for coefficient 'a'.")
        self.a = a
        return self

    @validate_inputs({"b": "positive"})
    def set_b(
        self, b: float, b_units: str = "K"
    ) -> "AntoineVaporPressureBuilder":
        """Set the coefficient 'b' of the Antoine equation."""
        if b_units == "K":
            self.b = b
            return self
        raise ValueError("Only K units are supported for coefficient 'b'.")

    @validate_inputs({"c": "positive"})
    def set_c(
        self, c: float, c_units: str = "K"
    ) -> "AntoineVaporPressureBuilder":
        """Set the coefficient 'c' of the Antoine equation."""
        if c_units == "K":
            self.c = c
            return self
        raise ValueError("Only K units are supported for coefficient 'c'.")

    def build(self) -> AntoineVaporPressureStrategy:
        """Validate and return an AntoineVaporPressureStrategy using the set
        coefficients.

        Returns:
            - Configured with coefficients a, b, and c.
        """
        self.pre_build_check()
        return AntoineVaporPressureStrategy(self.a, self.b, self.c)


class ClausiusClapeyronVaporPressureBuilder(BuilderABC):
    """Builder class for ClausiusClapeyronStrategy. This class facilitates
    setting the latent heat of vaporization, initial temperature, and initial
    pressure with unit handling and then builds the strategy object.

    The Clausius–Clapeyron relation can be approximated as:

    - dP / dT = (L / (R × T²))


    Methods:
    - set_latent_heat : Set latent heat in J/mol (or convertible units).
    - set_temperature_initial : Set initial temperature in K
        (or convertible units).
    - set_pressure_initial : Set initial pressure in Pa
        (or convertible units).
    - build : Validate parameters and return a ClausiusClapeyronStrategy.


    Example:
        ``` py title="ClausiusClapeyronVaporPressureBuilder"
        strategy = (
            ClausiusClapeyronVaporPressureBuilder()
            .set_latent_heat(2260)
            .set_temperature_initial(373.15)
            .set_pressure_initial(101325)
            .build()
        )
        ```

        ``` py title="ClausiusClapeyronVaporPressureBuilder with units"
        strategy = (
            ClausiusClapeyronVaporPressureBuilder()
            .set_latent_heat(2260, "J/mol")
            .set_temperature_initial(373.15, "K")
            .set_pressure_initial(101325, "Pa")
            .build()
        )
        ```

    References:
        - Equation: dP/dT = L / (R * T^2)
          https://en.wikipedia.org/wiki/Clausius%E2%80%93Clapeyron_relation
    """

    def __init__(self):
        """Initialize the Clausius-Clapeyron vapor pressure builder.

        Sets up the builder with required parameters for creating a
        ClausiusClapeyronStrategy, including latent heat, initial
        temperature, and initial pressure.
        """
        required_keys = [
            "latent_heat",
            "temperature_initial",
            "pressure_initial",
        ]
        super().__init__(required_keys)
        self.latent_heat = None
        self.temperature_initial = None
        self.pressure_initial = None

    @validate_inputs({"latent_heat": "positive"})
    def set_latent_heat(
        self, latent_heat: float, latent_heat_units: str
    ) -> "ClausiusClapeyronVaporPressureBuilder":
        """Set the latent heat of vaporization: Default units J/mol."""
        if latent_heat_units == "J/mol":
            self.latent_heat = latent_heat
            return self
        self.latent_heat = latent_heat * get_unit_conversion(
            latent_heat_units, "J/mol"
        )
        return self

    @validate_inputs({"temperature_initial": "positive"})
    def set_temperature_initial(
        self, temperature_initial: float, temperature_initial_units: str
    ) -> "ClausiusClapeyronVaporPressureBuilder":
        """Set the initial temperature. Default units: K."""
        if temperature_initial_units == "K":
            self.temperature_initial = temperature_initial
            return self
        self.temperature_initial = get_unit_conversion(
            temperature_initial_units, "K", temperature_initial
        )
        return self

    @validate_inputs({"pressure_initial": "positive"})
    def set_pressure_initial(
        self, pressure_initial: float, pressure_initial_units: str
    ) -> "ClausiusClapeyronVaporPressureBuilder":
        """Set the initial pressure. Default units: Pa."""
        if pressure_initial_units == "Pa":
            self.pressure_initial = pressure_initial
            return self
        self.pressure_initial = pressure_initial * get_unit_conversion(
            pressure_initial_units, "Pa"
        )
        return self

    def build(self) -> ClausiusClapeyronStrategy:
        """Validate parameters and return a ClausiusClapeyronStrategy object.

        Returns:
            - Configured with latent heat, initial Temperature, and Pressure.
        """
        self.pre_build_check()
        return ClausiusClapeyronStrategy(
            self.latent_heat,  # type: ignore
            self.temperature_initial,  # type: ignore
            self.pressure_initial,  # type: ignore
        )


class ConstantVaporPressureBuilder(BuilderABC):
    """Builder class for ConstantVaporPressureStrategy. This class facilitates
    setting the constant vapor pressure and then building the strategy object.

    Attributes:
        - vapor_pressure: The vapor pressure in Pa (scalar/float).

    Methods:
    - set_vapor_pressure: Set the constant vapor pressure in Pa
      (or convertible units).
    - build : Validate parameters and return a ConstantVaporPressureStrategy.

    Example:
        ``` py title="ConstantVaporPressureBuilder"
        strategy = (
            ConstantVaporPressureBuilder()
            .set_vapor_pressure(101325)
            .build()
        )
        ```

        ``` py title="ConstantVaporPressureBuilder with units"
        strategy = (
            ConstantVaporPressureBuilder()
            .set_vapor_pressure(1, "atm")
            .build()
        )
        ```

    References:
        - Equation: P = vapor_pressure
          https://en.wikipedia.org/wiki/Vapor_pressure
    """

    def __init__(self):
        """Initialize the Constant vapor pressure builder.

        Sets up the builder with the required vapor_pressure parameter.
        """
        required_keys = ["vapor_pressure"]
        super().__init__(required_keys)
        self.vapor_pressure = None

    @validate_inputs({"vapor_pressure": "positive"})
    def set_vapor_pressure(
        self, vapor_pressure: float, vapor_pressure_units: str
    ) -> "ConstantVaporPressureBuilder":
        """Set the constant vapor pressure."""
        if vapor_pressure_units == "Pa":
            self.vapor_pressure = vapor_pressure
            return self
        self.vapor_pressure = vapor_pressure * get_unit_conversion(
            vapor_pressure_units, "Pa"
        )
        return self

    def build(self) -> ConstantVaporPressureStrategy:
        """Validate parameters and return ConstantVaporPressureStrategy.

        Returns:
            - Configured with vapor_pressure in Pa.
        """
        self.pre_build_check()
        return ConstantVaporPressureStrategy(self.vapor_pressure)


class SaturationConcentrationVaporPressureBuilder(
    BuilderABC,
    BuilderMolarMassMixin,
    BuilderTemperatureMixin,
):
    """Builder class for ConstantVaporPressureStrategy.

    It allows for setting the vapor pressure using the saturation
    concentration, molar mass, and temperature. The saturation concentration is
    commonly called C^sat (or C* when in a mixture) in aerosol sciences.

    Methods:
    - set_saturation_concentration : Set the saturation concentration in
      kg/m³ (or convertible units).
    - set_molar_mass : Set the molar mass in kg/mol (or convertible units).
    - set_temperature : Set the temperature in K (or convertible units).
    - build : Validate parameters and return a ConstantVaporPressureStrategy.

    Example:
        ```py title="SaturationConcentrationVaporPressureBuilder"
        import particula as par
        strategy = (
            par.gas.SaturationConcentrationVaporPressureBuilder()
            .set_saturation_concentration(10e-6, "kg/m^3")
            .set_molar_mass(0.200, "kg/mol")
            .set_temperature(298.15, "K")
            .build()
        )
        ```

    References:
    - Donahue, N. M., Robinson, A. L., Stanier, C. O., & Pandis, S. N. (2006).
      Coupled Partitioning, Dilution, and Chemical Aging of Semivolatile
      Organics. Environmental Science & Technology, 40(8), 2635–2643.
      [DOI](https://doi.org/10.1021/es052297c)
    """

    def __init__(self):
        """Initialize the Saturation Concentration vapor pressure builder.

        Sets up the builder with required parameters for creating a
        ConstantVaporPressureStrategy from saturation concentration,
        molar mass, and temperature.
        """
        required_parameters = [
            "saturation_concentration",
            "molar_mass",
            "temperature",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderMolarMassMixin.__init__(self)
        BuilderTemperatureMixin.__init__(self)
        self.saturation_concentration = None
        self.molar_mass = None
        self.temperature = None

    @validate_inputs({"saturation_concentration": "nonnegative"})
    def set_saturation_concentration(
        self,
        saturation_concentration: float,
        saturation_concentration_units: str,
    ) -> "SaturationConcentrationVaporPressureBuilder":
        """Set the saturation concentration (C*, C^sat).

        Arguments:
            - saturation_concentration : Value of the saturation concentration.
            - saturation_concentration_units : Any units convertible to
                ``"kg/m^3"`` via ``get_unit_conversion`` are accepted.

        Returns:
           - The builder itself for fluent chaining.

        Example:
            ```py title="SaturationConcentrationVaporPressureBuilder"
            import particula as par
            strategy = (
                par.gas.SaturationConcentrationVaporPressureBuilder()
                .set_saturation_concentration(10e-6, "kg/m^3")
            )
            ```
        """
        if saturation_concentration_units == "kg/m^3":
            self.saturation_concentration = saturation_concentration
            return self
        self.saturation_concentration = (
            saturation_concentration
            * get_unit_conversion(saturation_concentration_units, "kg/m^3")
        )
        return self

    def build(self) -> ConstantVaporPressureStrategy:
        """Validate all parameters and construct a
        ConstantVaporPressureStrategy whose value is obtained from the
        ideal-gas relationship between concentration and partial pressure.

        Returns:
            ConstantVaporPressureStrategy containing the calculated vapor
            pressure.
        """
        self.pre_build_check()

        vapor_pressure = get_partial_pressure(
            concentration=self.saturation_concentration,
            molar_mass=self.molar_mass,
            temperature=self.temperature,
        )
        return ConstantVaporPressureStrategy(vapor_pressure)


class WaterBuckVaporPressureBuilder(BuilderABC):  # pylint: disable=too-few-public-methods
    """Builder class for WaterBuckStrategy.

    This class facilitates the building of the WaterBuckStrategy object.
    Which as of now has no additional parameters to set but could be
    extended in the future (e.g., ice-only calculations).

    Example:
        ```py title="WaterBuckVaporPressureBuilder"
        import particula as par
        strategy = par.gas.WaterBuckVaporPressureBuilder().build()
        ```
    """

    def __init__(self):
        """Initialize the Water Buck vapor pressure builder.

        Sets up the builder for creating a WaterBuckStrategy.
        """
        super().__init__()

    def build(self) -> WaterBuckStrategy:
        """Build and return a WaterBuckStrategy object.

        Returns:
            - Configured for water-specific Buck vapor pressure.
        """
        return WaterBuckStrategy()


class TableVaporPressureBuilder(
    BuilderABC,
    BuilderTemperatureTableMixin,
):
    """Builder for TableVaporPressureStrategy.

    Allows setting lookup tables of vapor pressure and temperature, handling
    optional unit conversion before constructing the strategy.
    """

    def __init__(self) -> None:
        """Initialize the Table vapor pressure builder.

        Sets up the builder with required parameters for creating a
        TableVaporPressureStrategy from lookup tables of vapor pressure
        and temperature.
        """
        required_parameters = ["vapor_pressure_table", "temperature_table"]
        BuilderABC.__init__(self, required_parameters)
        BuilderTemperatureTableMixin.__init__(self)
        self.vapor_pressure_table: NDArray[np.float64] | None = None

    @validate_inputs({"vapor_pressure_table": "positive"})
    def set_vapor_pressure_table(
        self,
        vapor_pressure_table: NDArray[np.float64],
        vapor_pressure_table_units: str = "Pa",
    ) -> "TableVaporPressureBuilder":
        """Set the vapor pressure lookup table."""
        table = np.asarray(vapor_pressure_table, dtype=np.float64)
        if vapor_pressure_table_units != "Pa":
            table = table * get_unit_conversion(
                vapor_pressure_table_units,
                "Pa",
            )
        self.vapor_pressure_table = table
        return self

    def build(self) -> TableVaporPressureStrategy:
        """Validate and construct a TableVaporPressureStrategy."""
        self.pre_build_check()
        return TableVaporPressureStrategy(
            vapor_pressures=self.vapor_pressure_table,  # type: ignore
            temperatures=self.temperature_table,
        )



"""Tests for the TurbulentShearCoagulationBuilder class."""

# pylint: disable=duplicate-code

import pytest

# pylint: disable=line-too-long
from particula.dynamics import (
    TurbulentShearCoagulationBuilder,
    TurbulentShearCoagulationStrategy,
)


def test_build_with_valid_parameters():
    """Test that building with valid parameters returns a
    TurbulentShearCoagulationStrategy.
    """
    builder = TurbulentShearCoagulationBuilder()
    builder.set_distribution_type("discrete")
    builder.set_turbulent_dissipation(
        1e-4, turbulent_dissipation_units="m^2/s^3"
    )
    builder.set_fluid_density(1.2, fluid_density_units="kg/m^3")
    strategy = builder.build()
    assert isinstance(strategy, TurbulentShearCoagulationStrategy)


def test_build_missing_required_parameters():
    """Test that building without required parameters raises a ValueError."""
    builder = TurbulentShearCoagulationBuilder()

    with pytest.raises(ValueError):
        builder.build()

    builder.set_distribution_type("discrete")
    with pytest.raises(ValueError):
        builder.build()

    builder.set_turbulent_dissipation(
        1e-4, turbulent_dissipation_units="m^2/s^3"
    )
    with pytest.raises(ValueError):
        builder.build()

    builder.set_fluid_density(1.2, fluid_density_units="kg/m^3")
    # With all required parameters set, it should now succeed:
    strategy = builder.build()
    assert isinstance(strategy, TurbulentShearCoagulationStrategy)



"""Sedimentation Coagulation builder class."""

from particula.abc_builder import BuilderABC
from particula.dynamics.coagulation.coagulation_builder import (
    coagulation_builder_mixin,
)

from ..coagulation_strategy.coagulation_strategy_abc import (
    CoagulationStrategyABC,
)
from ..coagulation_strategy.sedimentation_coagulation_strategy import (
    SedimentationCoagulationStrategy,
)


class SedimentationCoagulationBuilder(
    BuilderABC,
    coagulation_builder_mixin.BuilderDistributionTypeMixin,
):
    """Sedimentation Coagulation builder class.

    Creates a `SedimentationCoagulationStrategy` given a distribution type
    (e.g., "discrete", "continuous_pdf", or "particle_resolved"). Ensures
    the required parameters are set before building the strategy.

    Attributes:
        - distribution_type : Representation of the particle
          size distribution.

    Methods:
    - set_distribution_type : Set the distribution type.
    - build : Validate parameters and return a
      SedimentationCoagulationStrategy instance.

    Examples:
        ```py title="Example of using SedimentationCoagulationBuilder"
        import particula as par
        builder = SedimentationCoagulationBuilder()
        builder.set_distribution_type("discrete")
        strategy = builder.build()
        # strategy is now a SedimentationCoagulationStrategy instance
        ```
    """

    def __init__(self):
        """Initialize the Sedimentation coagulation builder.

        Sets up the builder with required parameters for creating a
        SedimentationCoagulationStrategy.
        """
        required_parameters = ["distribution_type"]
        BuilderABC.__init__(self, required_parameters)
        coagulation_builder_mixin.BuilderDistributionTypeMixin.__init__(self)

    def build(self) -> CoagulationStrategyABC:
        """Validate and return SedimentationCoagulationStrategy object."""
        self.pre_build_check()
        return SedimentationCoagulationStrategy(
            distribution_type=self.distribution_type,
        )



"""Unit tests for particula.particles.representation.get_sorted_bins_by_radius.

The tests cover:
• Already-sorted 1-D inputs (function must return inputs unchanged).
• Unsorted 1-D inputs (function must reorder accompanying arrays).
• Unsorted inputs with a 2-D distribution matrix (rows must be reordered
  consistently with the radii).
• Unsorted inputs with a 2-D distribution matrix and an all-zero charge array.
"""

import numpy as np

from particula.particles.representation import get_sorted_bins_by_radius


def test_sorted_vector():
    """Already-sorted 1-D input should remain unchanged."""
    radius = np.array([1.0, 2.0, 3.0])
    dist = np.array([10.0, 20.0, 30.0])
    conc = np.array([100, 200, 300])
    charge = np.array([1, 2, 3])

    d2, c2, q2 = get_sorted_bins_by_radius(radius, dist, conc, charge)

    np.testing.assert_array_equal(d2, dist)
    np.testing.assert_array_equal(c2, conc)
    np.testing.assert_array_equal(q2, charge)


def test_unsorted_vector():
    """Unsorted 1-D input must be reordered into ascending radii."""
    radius = np.array([2.0, 1.0, 3.0])
    dist = np.array([20.0, 10.0, 30.0])
    conc = np.array([200, 100, 300])
    charge = np.array([2, 1, 3])

    d2, c2, q2 = get_sorted_bins_by_radius(radius, dist, conc, charge)

    np.testing.assert_array_equal(d2, np.array([10.0, 20.0, 30.0]))
    np.testing.assert_array_equal(c2, np.array([100, 200, 300]))
    np.testing.assert_array_equal(q2, np.array([1, 2, 3]))


def test_unsorted_matrix():
    """Unsorted radii with 2-D distribution; rows must be reordered."""
    radius = np.array([2.0, 1.0, 3.0])
    dist = np.array([[20.0, 21.0], [10.0, 11.0], [30.0, 31.0]])
    conc = np.array([200, 100, 300])
    charge = np.array([2, 1, 3])

    d2, c2, q2 = get_sorted_bins_by_radius(radius, dist, conc, charge)

    np.testing.assert_array_equal(
        d2, np.array([[10.0, 11.0], [20.0, 21.0], [30.0, 31.0]])
    )
    np.testing.assert_array_equal(c2, np.array([100, 200, 300]))
    np.testing.assert_array_equal(q2, np.array([1, 2, 3]))


def test_unsorted_matrix_zero_charge():
    """Unsorted radii with 2-D distribution and zero-charge array."""
    radius = np.array([3.0, 1.0, 2.0])
    dist = np.array([[30.0, 31.0], [10.0, 11.0], [20.0, 21.0]])
    conc = np.array([300, 100, 200])
    charge = np.array([0])  # all zeros

    d2, c2, q2 = get_sorted_bins_by_radius(radius, dist, conc, charge)

    np.testing.assert_array_equal(
        d2, np.array([[10.0, 11.0], [20.0, 21.0], [30.0, 31.0]])
    )
    np.testing.assert_array_equal(c2, np.array([100, 200, 300]))
    np.testing.assert_array_equal(q2, charge)  # still all zeros



"""Calculates the normalized acceleration variance in isotropic turbulence."""

from typing import Union

import numpy as np
from numpy.typing import NDArray

from particula.util.validate_inputs import validate_inputs


@validate_inputs({"re_lambda": "positive"})
def get_normalized_accel_variance_ao2008(
    re_lambda: Union[float, NDArray[np.float64]],
    numerical_stability_epsilon: float = 1e-14,
) -> Union[float, NDArray[np.float64]]:
    """Calculate the normalized acceleration variance in isotropic turbulence.

    This coefficient describes the statistical behavior of acceleration
    fluctuations in turbulent flows.

    - a_o = (11 + 7 R_λ) / (205 + R_λ)
        - a_o is Normalized acceleration variance in isotropic turbulence [-].
        - R_λ is Taylor-microscale Reynolds number [-].

    Where:
        - a_o (accel_variance) is Normalized acceleration variance in isotropic
          turbulence [-].
        - R_λ (re_lambda) is Taylor-microscale Reynolds number [-].
        - ε (numerical_stability_epsilon) is Small number added to R_λ
          for numerical stability.

    Arguments:
        - re_lambda : Taylor-microscale Reynolds number [-]

    Returns:
        - accel_variance : Normalized acceleration variance [-]

    Examples:
        ```py title="Example Usage"
        import particula as par
        par.gas.get_normalized_accel_variance_ao2008(500.0)
        # Output: ~0.05
        ```

    References:
    - The equivalent numerically stable version used is this.
        (7 + 11 / (R_λ + ε)) / (1 + 205 / (R_λ + ε))
    - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
        the geometric collision rate of sedimenting droplets. Part 2. Theory
        and parameterization. New Journal of Physics, 10.
        https://doi.org/10.1088/1367-2630/10/7/075016
    """
    return (7 + 11 / (re_lambda + numerical_stability_epsilon)) / (
        1 + 205 / (re_lambda + numerical_stability_epsilon)
    )



"""Tests for the BrownianCoagulationBuilder class."""

import pytest

from particula.dynamics import (
    BrownianCoagulationBuilder,
    BrownianCoagulationStrategy,
)


def test_build_with_valid_parameters():
    """Test that building with valid parameters returns."""
    builder = BrownianCoagulationBuilder()
    builder.set_distribution_type("discrete")
    strategy = builder.build()
    assert isinstance(strategy, BrownianCoagulationStrategy)


def test_build_missing_required_parameters():
    """Test that building without required parameters raises ValueError."""
    builder = BrownianCoagulationBuilder()
    with pytest.raises(ValueError):
        builder.build()
    with pytest.raises(ValueError):
        builder.set_distribution_type("NotRight")



"""Tests for the VaporPressureBuilder class."""

import pytest

from particula.gas.vapor_pressure_builders import (
    AntoineVaporPressureBuilder,
    ClausiusClapeyronVaporPressureBuilder,
    ConstantVaporPressureBuilder,
    SaturationConcentrationVaporPressureBuilder,
    TableVaporPressureBuilder,
    WaterBuckVaporPressureBuilder,
)
from particula.gas.vapor_pressure_strategies import TableVaporPressureStrategy


def test_antoine_set_positive_a():
    """Test setting a positive value for the coefficient 'a'."""
    builder = AntoineVaporPressureBuilder()
    assert builder.set_a(10).a == 10


def test_antoine_set_negative_a():
    """Test setting a negative value for the coefficient 'a'."""
    builder = AntoineVaporPressureBuilder()
    with pytest.raises(ValueError) as excinfo:
        builder.set_a(-5)
    assert "Coefficient 'a' must be a positive value." in str(excinfo.value)


def test_antoine_set_parameters_with_missing_key():
    """Test setting parameters with a missing key in the dictionary."""
    builder = AntoineVaporPressureBuilder()
    with pytest.raises(ValueError) as excinfo:
        builder.set_parameters({"a": 10, "b": 2000})
    assert "Missing required parameter(s): c" in str(excinfo.value)


def test_antoine_build_without_all_coefficients():
    """Test building the strategy without all coefficients set."""
    builder = AntoineVaporPressureBuilder().set_a(10).set_b(2000)
    with pytest.raises(ValueError) as excinfo:
        builder.build()
    assert "Required parameter(s) not set: c" in str(excinfo.value)


def test_clausius_set_latent_heat_positive():
    """Test setting a positive value for the latent heat."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    builder.set_latent_heat(100, "J/mol")
    assert builder.latent_heat == 100, "Latent heat should be set correctly"


def test_clausius_set_latent_heat_negative():
    """Test setting a negative value for the latent heat."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    with pytest.raises(ValueError):
        builder.set_latent_heat(-100, "J/mol")


def test_clausius_set_temperature_initial_positive():
    """Test setting a positive value for the initial temperature."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    builder.set_temperature_initial(300, "K")
    assert builder.temperature_initial == 300


def test_clausius_set_temperature_initial_negative():
    """Test setting a negative value for the initial temperature."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    with pytest.raises(ValueError):
        builder.set_temperature_initial(-5, "K")


def test_clausius_set_pressure_initial_positive():
    """Test setting a positive value for the initial pressure."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    builder.set_pressure_initial(101325, "Pa")
    assert builder.pressure_initial == 101325


def test_clausius_set_pressure_initial_negative():
    """Test setting a negative value for the initial pressure."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    with pytest.raises(ValueError):
        builder.set_pressure_initial(-101325, "Pa")


def test_clausius_set_parameters_complete():
    """Test setting all parameters at once, with different pressure units."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    parameters = {
        "latent_heat": 2260,
        "latent_heat_units": "J/mol",
        "temperature_initial": 373,
        "temperature_initial_units": "K",
        "pressure_initial": 101325,
        "pressure_initial_units": "Pa",
    }
    builder.set_parameters(parameters)
    assert builder.latent_heat == 2260
    assert builder.temperature_initial == 373
    assert builder.pressure_initial == 101325


def test_clausius_build_success():
    """Test building the strategy successfully."""
    builder = (
        ClausiusClapeyronVaporPressureBuilder()
        .set_latent_heat(2260, "J/mol")
        .set_temperature_initial(373, "K")
        .set_pressure_initial(101325, "Pa")
    )
    builder.build()


def test_clausius_build_failure():
    """Test building the strategy without all parameters set."""
    builder = ClausiusClapeyronVaporPressureBuilder()
    builder.set_latent_heat(2260, "J/mol").set_temperature_initial(373, "K")
    with pytest.raises(ValueError) as excinfo:
        builder.build()
    assert "Required parameter(s) not set: pressure_initial" in str(
        excinfo.value
    )


def test_constant_set_vapor_pressure_positive():
    """Test setting a positive value for the constant vapor pressure."""
    builder = ConstantVaporPressureBuilder()
    builder.set_vapor_pressure(101325, "Pa")
    assert builder.vapor_pressure == 101325


def test_constant_set_vapor_pressure_negative():
    """Test setting a negative value for the constant vapor pressure."""
    builder = ConstantVaporPressureBuilder()
    with pytest.raises(ValueError):
        builder.set_vapor_pressure(-101325, "Pa")


def test_constant_set_parameters_complete():
    """Test setting all parameters at once. With different pressure units."""
    builder = ConstantVaporPressureBuilder()
    parameters = {"vapor_pressure": 101325, "vapor_pressure_units": "Pa"}
    builder.set_parameters(parameters)
    assert builder.vapor_pressure == 101325


def test_constant_build_success():
    """Test building the strategy successfully."""
    builder = ConstantVaporPressureBuilder()
    builder.set_vapor_pressure(101325, "Pa")
    builder.build()


def test_constant_build_failure():
    """Test building the strategy without setting the vapor pressure."""
    builder = ConstantVaporPressureBuilder()
    with pytest.raises(ValueError) as excinfo:
        builder.build()
    assert "Required parameter(s) not set: vapor_pressure" in str(excinfo.value)


def test_build_water_buck():
    """Test building the WaterBuck strategy."""
    builder = WaterBuckVaporPressureBuilder()
    builder.build()


# ----------------------------------------------------------------------
# SaturationConcentrationVaporPressureBuilder tests
# ----------------------------------------------------------------------


def test_saturation_set_concentration_positive():
    """Test setting a positive saturation concentration (default units)."""
    builder = SaturationConcentrationVaporPressureBuilder()
    builder.set_saturation_concentration(1e-6, "kg/m^3")
    assert builder.saturation_concentration == pytest.approx(1e-6)


def test_saturation_build_success():
    """Test successful build when all parameters are provided."""
    builder = (
        SaturationConcentrationVaporPressureBuilder()
        .set_saturation_concentration(1e-6, "kg/m^3")
        .set_molar_mass(0.018, "kg/mol")
        .set_temperature(298.15, "K")
    )
    builder.build()  # Should not raise


def test_saturation_build_failure_missing_param():
    """Test build failure when temperature is missing."""
    builder = (
        SaturationConcentrationVaporPressureBuilder()
        .set_saturation_concentration(1e-6, "kg/m^3")
        .set_molar_mass(0.018, "kg/mol")
    )
    with pytest.raises(ValueError) as excinfo:
        builder.build()
    assert "Required parameter(s) not set: temperature" in str(excinfo.value)


def test_table_builder_set_table_and_build():
    """Test building TableVaporPressureStrategy with tables."""
    builder = (
        TableVaporPressureBuilder()
        .set_vapor_pressure_table([1000.0, 2000.0], "Pa")
        .set_temperature_table([300.0, 350.0], "K")
    )
    strategy = builder.build()
    assert isinstance(strategy, TableVaporPressureStrategy)


def test_table_builder_missing_temperature_table():
    """Ensure error raised when temperature table not set."""
    builder = TableVaporPressureBuilder().set_vapor_pressure_table(
        [1000.0, 2000.0], "Pa"
    )
    with pytest.raises(ValueError):
        builder.build()
